# Copyright© 1986-2024 Altair Engineering Inc.

"""Utilities for logging on TTY"""

import sys

_DEBUG = False


def set_debug(dbg=False):
    """Activate debug mode"""
    # pylint: disable=global-statement
    global _DEBUG
    _DEBUG = dbg


def write(msg, add_return=True, error=False):
    """Print the `msg` to the stdout"""
    if add_return:
        msg = str(msg) + "\n"
    fd = sys.stdout
    if error:
        fd = sys.stderr
    fd.write(msg)
    fd.flush()


def debug(msg):
    """Print the `msg` to the stdout if debug mode is set"""
    if _DEBUG:
        write(msg, error=True)



# Copyright© 1986-2024 Altair Engineering Inc.

"""Module containing extensions for pkr"""
import abc
from builtins import str
import signal
import pkgutil

try:
    from importlib.metadata import entry_points
except ModuleNotFoundError:
    from importlib_metadata import entry_points

from pkr.cli.log import write
from pkr.utils import PkrException, get_pkr_path


class ExtMixin(metaclass=abc.ABCMeta):
    """Mixin for an extension implementation"""

    @staticmethod
    def setup(args, kard):
        """Populate build_data with extension specific values.

        This hook is called when a new kard is created. It might also be used
        for kard updating.

        A practical implementation is provided in the function
        `ensure_definition_matches` in utils.

        Args:
          - args: the args passed in the env
          - kard: the kard object (which might not be available through
          load_current when a new kard is created)
        """

    @staticmethod
    def get_context_template_data(password=None) -> dict:
        """Return a map that will be added to the template data"""

    @staticmethod
    def post_up(effective_modules):
        """Run specific post up jobs"""

    @staticmethod
    def populate_kard():
        """Populate kard folder"""

    @staticmethod
    def configure_parser(parser):
        """configure option parser"""


class Extensions:
    """This class allow the recursive call of a specific method on all
    extensions
    """

    def __init__(self, features=None):
        super().__init__()
        self.features = features
        self._extensions = None

    @property
    def extensions(self):
        """Lazy loading of extensions"""
        if self._extensions is None:
            self._extensions = self.list_all()
        if self.features is not None:
            return {
                name: self._extensions[name] for name in self.features if name in self._extensions
            }
        return self._extensions

    def __getattr__(self, attribute):
        """Return all matching extensions methods wrapped into a lambda"""
        # Unknown extension method
        if not hasattr(ExtMixin, attribute):
            return super().__getattribute__(attribute)

        # Called outside a kard, doing nothing
        if not self.features:
            return lambda *args, **kw: ()

        def wrapper(*args, **kwargs):
            return [
                output[1]
                for output in map(
                    lambda ext: self._wrap_call(ext[0], ext[1], attribute, *args, *kwargs),
                    self.extensions.items(),
                )
                if output is not None
            ]

        return wrapper

    @staticmethod
    def _wrap_call(name, extension, method_name, *args, **kwargs):
        method = getattr(extension, method_name, None)
        base_method = getattr(ExtMixin, method_name, None)

        if method is None or method is base_method:
            # it is OK if an extension does not implement all methods
            return None

        try:
            return (name, method(*args, **kwargs))
        except TimeoutException:
            write(f'Extension "{extension.name}" raise timeout error, step "{method_name}"')
            raise
        except PkrException:
            # If this is a PkrException, we simply propagate it, and delegate its handling to the
            # caller
            raise
        except Exception as exc:
            write(
                f'Extension "{extension.name}" raise an unknown exception, step "{method_name}": '
                f"{exc}"
            )
            raise

    def list(self):
        """Return the list of the extensions"""
        return self.extensions.keys()

    @classmethod
    def _get_extension_class(cls, module):
        for attr in dir(module):
            ext_cls = getattr(module, attr)
            try:
                if issubclass(ext_cls, ExtMixin) and ext_cls != ExtMixin:
                    return ext_cls
            except TypeError:
                pass
        return None

    @classmethod
    def list_all(cls):
        """Return the list of all available extensions"""
        # Load from pkr path
        extensions = {}
        for importer, package_name, _ in pkgutil.iter_modules(
            [str(get_pkr_path() / "extensions")]
        ):
            module = importer.find_spec(package_name).loader.load_module(package_name)
            extensions[package_name] = cls._get_extension_class(module)
        # Load from pkr_extensions entrypoints (and TO BE DEPRECATED extensions group)
        eps = entry_points()
        if hasattr(eps, "select"):
            extension_eps = list(eps.select(group="pkr_extensions")) + list(
                eps.select(group="extensions")
            )
        else:
            extension_eps = eps.get("pkr_extensions", ()) + eps.get(  # pylint: disable=no-member
                "extensions", ()
            )

        for entry in extension_eps:
            if entry.name not in extensions:
                extensions[entry.name] = entry.load()

        return extensions

    def __contains__(self, ext_name):
        return ext_name in self.features


class TimeoutException(Exception):
    """Exception for timeout"""


def timeout_handler(*_):
    """Simple function for raising a timeout Exception"""
    raise TimeoutException()


def timeout(timeout_duration):  # pylint: disable=C0111
    def wrap(func):  # pylint: disable=C0111
        def decorated_function(*args, **kwargs):  # pylint: disable=C0111
            signal.signal(signal.SIGALRM, timeout_handler)
            signal.alarm(timeout_duration)
            ret = func(*args, **kwargs)
            signal.alarm(0)
            return ret

        return decorated_function

    return wrap



# Copyright© 1986-2024 Altair Engineering Inc.

"""CLI argument extend action"""

from argparse import Action, OPTIONAL
import copy


# pylint: disable=too-many-arguments,redefined-builtin
class ExtendAction(Action):
    """CLI argument extend action"""

    def __init__(
        self,
        option_strings,
        dest,
        nargs=None,
        const=None,
        default=None,
        type=None,
        choices=None,
        required=False,
        help=None,
        metavar=None,
    ):
        if nargs == 0:
            raise ValueError(
                "nargs for append actions must be != 0; if arg string are not "
                "supplying the value to append, the append const action may "
                "be more appropriate"
            )
        if const is not None and nargs != OPTIONAL:
            raise ValueError(f"nargs must be {OPTIONAL} to supply const")

        super().__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=nargs,
            const=const,
            default=default,
            type=type,
            choices=choices,
            required=required,
            help=help,
            metavar=metavar,
        )

    def __call__(self, parser, namespace, values, option_string=None):
        items = getattr(namespace, self.dest, None)

        if items is None:
            items = []
        elif isinstance(items, list):
            items = items[:]
        else:
            items = copy.copy(items)

        items.extend(values)
        setattr(namespace, self.dest, items)



# Copyright© 1986-2024 Altair Engineering Inc.

"""pkr functions for creating the context"""

from collections import namedtuple
from concurrent.futures import ThreadPoolExecutor
import os
import re
import sys
import traceback
from pathlib import Path

import docker
import tenacity

from pkr.driver import _USE_ENV_VAR
from pkr.driver.base import AbstractDriver
from pkr.cli.log import write
from pkr.utils import PkrException

DOCKER_SOCK = "unix://var/run/docker.sock"
DOCKER_CLIENT_TIMEOUT = int(os.environ.get("DOCKER_CLIENT_TIMEOUT", 300))


class ImagePullError(PkrException):
    """Raise when error occurs while pulling image"""


class DockerRegistry(namedtuple("DockerRegistry", ("url", "username", "password"))):
    """A Docker registry representation

    Args:
      * url: the URL of the registry
      * username: the username for authenticating on the registry
      * password: the password for authenticating on the registry
    """


# pylint: disable=abstract-method
class DockerDriver(AbstractDriver):
    """Context object used to manipulate the context used to generate
    docker images
    """

    DOCKER_CONTEXT = "docker-context"
    DOCKER_CONTEXT_SOURCE = "dockerfiles"

    def __init__(self, kard, password=None, **kwargs):
        super().__init__(kard=kard, password=password, **kwargs)
        self.metas = {"tag": None}
        # Both of these options work with APIClient and from_env
        kwargs.setdefault("timeout", DOCKER_CLIENT_TIMEOUT)
        kwargs.setdefault("version", "auto")
        if _USE_ENV_VAR:
            self.docker = docker.from_env(**kwargs).api
        else:
            self.docker = docker.APIClient(**kwargs)
        self.platform = os.environ.get("DOCKER_DEFAULT_PLATFORM")

    def get_meta(self, extras, kard):
        values = super().get_meta(extras, kard)
        if "tag" in extras:
            extras["tag"] = str(values["tag"])
        return values

    def get_templates(self):
        """Use the environment information to return files and folder to template
        or copy to templated directory according to the 'requires' section of the
        containers description
        """
        templates = []
        templates_path = (
            self.kard.env.pkr_path / self.kard.env.template_dir / self.DOCKER_CONTEXT_SOURCE
        )

        # Process dockerfiles
        for container in self.kard.env.get_container():
            context = self.kard.env.get_container(container).get("context", self.DOCKER_CONTEXT)

            # Process requirements
            for src in self.kard.env.get_requires([container]):
                template = {
                    "source": src["origin"],
                    "origin": src["origin"],
                    "destination": src["dst"],
                    "subfolder": context,
                    "excluded_paths": src.get("exclude", []),
                    "gen_template": False,
                }
                if template not in templates:
                    # Dedup templates to avoid multi-copy
                    templates.append(template)

            try:
                dockerfile = self.kard.env.get_container(container)["dockerfile"]
            except KeyError:
                # In this case, we use an image provided by the hub
                continue

            # Automatically add dockerfile name matching folder to the context
            dockerfile = Path(dockerfile).stem
            templates.append(
                {
                    "source": templates_path / f"{dockerfile}",  # Match template
                    "origin": templates_path,
                    "destination": "",
                    "subfolder": context,
                }
            )

            # Also add anything matching dockerfile name stem with an extension
            # suffix of some kind
            templates.append(
                {
                    "source": templates_path / f"{dockerfile}.*",  # Match template
                    "origin": templates_path,
                    "destination": "",
                    "subfolder": context,
                }
            )

        return templates

    def context_path(self, sub_path, container):
        """Return absolute path for sub_path relative to container context"""
        context = self.kard.env.get_container(container).get("context", self.DOCKER_CONTEXT)
        context_path = self.kard.real_path / context
        return context_path / sub_path

    def get_registry(self, **kwargs):
        """Return a DockerRegistry instance with either the provided values, or
        those present in the meta.
        """
        for var in ("url", "username", "password"):
            if var not in kwargs and var in self.kard.meta:
                kwargs[var] = self.kard.meta[var]

        return DockerRegistry(**kwargs)

    # pylint: disable=arguments-differ,too-many-arguments,too-many-locals
    def build_images(
        self,
        services,
        rebuild_context,
        tag=None,
        verbose=True,
        logfile=None,
        nocache=False,
        parallel=None,
        no_rebuild=False,
        target=None,
        **kwargs,
    ):
        """Build docker images.

        Args:
          * services: the name of the images to build
          * tag: the tag on which the image will be saved
          * verbose: verbose logs
          * logfile: separate log file for the underlying build
          * nocache: disable docker cache
          * parallel: (int|None) Number of concurrent build
          * no_rebuild: do not build if destination image exists
          * target: name of the build-stage to build in a multi-stage Dockerfile
        """
        services = services or list(self.kard.env.get_container().keys())
        if rebuild_context:
            self.kard.make()

        tag = tag or self.kard.meta["tag"]

        with LogOutput(logfile) as logfh:
            if parallel:
                if len(services) > 1:
                    logfh.write(f"Building docker images using {parallel} threads ...\n")
                futures = []
                with ThreadPoolExecutor(max_workers=parallel) as executor:
                    for service in services:
                        futures.append(
                            executor.submit(
                                self._build_image,
                                service,
                                tag,
                                verbose,
                                logfile,
                                nocache,
                                no_rebuild,
                                True,
                                target,
                            )
                        )
                for future in futures:
                    future.result(timeout=1800)
            else:
                if len(services) > 1:
                    logfh.write("Building docker images...\n")
                for service in services:
                    self._build_image(
                        service, tag, verbose, logfile, nocache, no_rebuild, False, target
                    )

    # pylint: disable=too-many-arguments
    def _build_image(
        self,
        service,
        tag=None,
        verbose=True,
        logfile=None,
        nocache=False,
        no_rebuild=False,
        bufferize=None,
        target=None,
    ):
        """Build docker image.

        Args:
          * service: service to build
          * tag: the tag on which the image will be saved
          * verbose: verbose logs
          * logfile: separate log file for the underlying build
          * nocache: disable docker cache
          * parallel: (int|None) Number of concurrent build
          * no_rebuild: do not build if destination image exists
          * target: name of the build-stage to build in a multi-stage Dockerfile
        """
        image_name = self.make_image_name(service, tag)

        with LogOutput(logfile, bufferize=bufferize) as logfh:
            dockerfile = self.kard.env.get_container(service).get("dockerfile")
            if not dockerfile:
                return
            if not target:
                target = self.kard.env.get_container(service).get("target")

            logfh.write(f"Building {image_name}{f'({target})' if target else ''} image...\n")

            if no_rebuild:
                image = len(self.docker.images(image_name)) == 1

            if not no_rebuild or image is False:
                context = self.kard.env.get_container(service).get("context", self.DOCKER_CONTEXT)
                stream = self.docker.build(
                    path=str(self.kard.path / context),
                    dockerfile=str(Path(self.kard.path / context, dockerfile)),  # Relative Path
                    tag=image_name,
                    decode=True,
                    nocache=nocache,
                    forcerm=True,
                    target=target,
                    platform=self.platform,
                )

                self.print_docker_stream(
                    stream, verbose=verbose, logfile=logfile, bufferize=bufferize
                )

            logfh.write("done.\n")

    def _logon_remote_registry(self, registry):
        """Push images to a remote registry

        Args:
          * registry: a DockerRegistry instance
        """
        if registry.username is None:
            return
        write(f"Logging to {registry.url}...")
        self.docker.login(
            username=registry.username, password=registry.password, registry=registry.url
        )

    # pylint: disable=arguments-differ,too-many-arguments,too-many-locals
    def push_images(
        self,
        services,
        registry,
        username,
        password,
        tag=None,
        other_tags=None,
        parallel=None,
        **kwargs,
    ):
        """Push images to a remote registry

        Args:
          * services: the name of the images to push
          * registry: a DockerRegistry instance
          * tag: the tag of the version to push
          * parallel: push parallelism
        """
        services = services or list(self.kard.env.get_container().keys())
        tag = tag or self.kard.meta["tag"]

        registry = self.get_registry(url=registry, username=username, password=password)
        if registry.username is not None:
            self._logon_remote_registry(registry)

        tags = [tag]
        tags.extend(other_tags)

        todos = []
        for service in services:
            image_name = self.make_image_name(service)
            image = self.make_image_name(service, tag)
            rep_tag = f"{registry.url}/{image_name}"
            todos.append((image, rep_tag, tags))

        if parallel:
            futures = []
            with ThreadPoolExecutor(max_workers=parallel) as executor:
                for todo in todos:
                    futures.append(executor.submit(self._push_image, *todo, buffer=True))
            for future in futures:
                future.result()
        else:
            for todo in todos:
                self._push_image(*todo)

    def _push_image(self, image, rep_tag, tags, buffer=False):
        """Push image to a remote registry

        Args:
          * image: the name of the image to push
          * rep_tag: distant image name
          * tags: list of tags to push to
        """
        for dest_tag in tags:
            if not buffer:
                write(f"Pushing {image} to {rep_tag}:{dest_tag}")
                sys.stdout.flush()

            try:
                self.docker.tag(image=image, repository=rep_tag, tag=dest_tag, force=True)

                ret = self.docker.push(repository=rep_tag, tag=dest_tag, decode=True, stream=True)

                error = ""
                for stream in ret:
                    if "error" in stream:
                        error += "\n" + stream["errorDetail"]["message"]

                if buffer:
                    write(f"Pushing {image} to {rep_tag}:{dest_tag}")
                    sys.stdout.flush()
                write(" Done !")
            except docker.errors.APIError as error:
                raise error

    def logon_remote_registry(self, registry, username=None, password=None):
        """Login to a remote registry

        Args:
          * registry: a DockerRegistry instance
          * username: username
          * password: password
        """
        registry = self.get_registry(url=registry, username=username, password=password)
        self._logon_remote_registry(registry)

    # pylint: disable=too-many-arguments,too-many-locals
    def pull_images(
        self,
        services,
        registry=None,
        username=None,
        password=None,
        tag=None,
        parallel=None,
        ignore_errors=False,
        **_,
    ):
        """Pull images from a remote registry

        Args:
          * services: the name of the images to pull
          * registry: a docker registry url
          * tag: the tag of the version to pull
          * parallel: pull parallelism
        """
        if registry is not None:
            services = services or list(self.kard.env.get_container().keys())
            remote_tag = tag or self.kard.meta["tag"]
            tag = self.kard.meta["tag"]

            docker_registry = self.get_registry(url=registry, username=username, password=password)
            self._logon_remote_registry(docker_registry)

            todos = []
            for service in services:
                image_name = self.make_image_name(service)
                image = self.make_image_name(service, tag)
                todos.append((image, image_name, docker_registry, remote_tag))
        else:
            todos = services
        if parallel:
            futures = []
            with ThreadPoolExecutor(max_workers=parallel) as executor:
                for image, image_name, reg, remote_tag in todos:
                    futures.append(
                        (
                            image,
                            executor.submit(
                                self._pull_image,
                                image_name,
                                reg.url,
                                tag,
                                remote_tag,
                                ignore_errors,
                            ),
                            image_name,
                            remote_tag,
                        )
                    )
            for image, future, image_name, remote_tag in futures:
                future.result()
                write(f"Pulling {image} from {registry.url}/{image_name}:{remote_tag}...")
                write(" Done !\n")
                sys.stdout.flush()
        else:
            for image, image_name, reg, remote_tag in todos:
                write(f"Pulling {image} from {reg.url}/{image_name}:{remote_tag}...")
                sys.stdout.flush()
                self._pull_image(image_name, reg.url, tag, remote_tag, ignore_errors)
                write(" Done !\n")

        write("All images have been pulled successfully !\n")

    def download_images(
        self, services, registry, username, password, tag=None, nopull=False, **kwargs
    ):
        """Download images from a remote registry and save to kard

        Args:
          * services: the name of the images to download
          * registry: a DockerRegistry instance
          * tag: the tag of the version to download
        """
        services = services or list(self.kard.env.get_container().keys())
        tag = tag or self.kard.meta["tag"]

        save_path = Path(self.kard.path) / "images"
        write(f"Cleaning images destination {save_path}")
        save_path.mkdir(exist_ok=True)
        for child in save_path.iterdir():
            child.unlink()

        if not nopull:
            self.pull_images(services, registry, username, password, tag=tag)

        for service in services:
            image_path = save_path / f"{service}.tar"
            image_name = self.make_image_name(service, tag)
            write(f"Saving {image_name} to {image_path}")
            sys.stdout.flush()

            with open(image_path, "wb") as f:
                for chunk in self.docker.get_image(image_name):
                    f.write(chunk)

            write(" Done !\n")
        write("All images have been saved successfully !\n")

    def import_images(self, services, **_):
        """Import images from kard to local docker

        Args:
          * services: the name of the images to load
          * tag: the tag of the version to load
        """
        services = services or list(self.kard.env.get_container().keys())

        save_path = Path(self.kard.path) / "images"
        for child in save_path.iterdir():
            service = child.name[:-4]
            if service not in services:
                continue
            write(f"Importing {child} ...")
            with open(child, "rb") as f:
                rsp = self.docker.load_image(f.read())
            for message in rsp:
                write(message.get("stream", ""))
            write("\n")
        write("All images have been loaded successfully !\n")

    @tenacity.retry(
        wait=tenacity.wait_fixed(1),
        stop=tenacity.stop_after_attempt(3),
        reraise=True,
        retry=tenacity.retry_if_exception_type(ImagePullError),
    )
    def _pull_image(self, image_name, registry_url, tag, remote_tag, ignore_errors):
        """
        Pull one image, retry few times to be robust to registry or network
        related issues.
        Usually, if an attempt fails, the next one will succeed.

        Args:
          * image_name: the name of the image to pull
          * registry_url: the DockerRegistry instance url
          * tag: the tag of the version to pull
        """

        rep_tag = f"{registry_url}/{image_name}"

        try:
            self.docker.pull(repository=rep_tag, tag=remote_tag)

            # Strip the repository tag
            self.docker.tag(
                image=":".join((rep_tag, remote_tag)), repository=image_name, tag=tag, force=True
            )

        except docker.errors.APIError as error:
            error_msg = f"Error while pulling the image {tag}: {error}"
            write(error_msg)
            if not ignore_errors:
                # pylint: disable=raise-missing-from
                raise ImagePullError(error_msg)

    @staticmethod
    def print_docker_stream(stream, verbose=True, logfile=None, bufferize=False):
        """Util method to print docker logs"""
        with LogOutput(logfile, bufferize=bufferize) as logfh:
            log_keys = set(("status", "stream"))
            all_logs = []
            last_log_id = [None]

            def print_log(log):
                for key in log_keys & set(log):
                    try:
                        if key == "status" and log.get(key) in ("Downloading", "Extracting"):
                            status_id = log.get("id")

                            if last_log_id[0] is None:
                                last_log_id[0] = status_id
                            if last_log_id[0] != status_id:
                                last_log_id[0] = status_id
                                logfh.writeln(log["progress"])
                            else:
                                logfh.write_console(log["progress"] + "\r")
                        else:
                            logfh.write_console("\n")
                            logfh.writeln(log.get(key))
                    except:
                        write(traceback.format_exc())
                        raise

            for log in stream:
                last_logs = []
                if log is None:
                    continue

                if isinstance(log, list):
                    last_logs.extend(log)
                    all_logs.extend(log)
                else:
                    last_logs.append(log)
                    all_logs.append(log)

                for last_log in last_logs:
                    if verbose:
                        print_log(last_log)

                    # Catch errors
                    if "error" in last_log:
                        for log_it in all_logs:
                            print_log(log_it)
                        raise Exception(
                            f"Error during docker process: {last_log['errorDetail']['message']}"
                        )

    def encrypt(self, password=None):
        """Hook for drivers to provide a kard encrypt feature"""
        raise NotImplementedError()

    def decrypt(self, password=None):
        """Hook for drivers to provide a kard decrypt feature"""
        raise NotImplementedError()

    def purge_images(self, tag=None, except_tag=None, repository=None, **kwargs):
        """Delete all images of this project.

        Only tag or except_tag can be specified simultaneously.

        Args:
          * except_tag: delete all image but this tag
          * tag: only delete this tag
          * repository: delete image reference in a specified repository
        """
        services = list(self.kard.env.get_container().keys())
        if except_tag is None:
            tag = tag or self.kard.meta["tag"]
        else:
            tag = f"(?!{except_tag})$"

        images_to_del = [self.make_image_name(s, tag) for s in services]

        if repository:
            tmp = []
            for image in images_to_del:
                tmp.append(image)
                tmp.append("/".join((repository, image)))
            images_to_del = tmp

        images_regex = "(" + ")|(".join(images_to_del) + ")"

        for img in self.docker.images():
            for repo_tag in img.get("RepoTags", []):
                if re.match(images_regex, repo_tag):
                    write("Deleting image " + repo_tag)
                    try:
                        self.docker.remove_image(repo_tag)
                    # pylint: disable=broad-exception-caught
                    except BaseException as exc:
                        write(exc)

    def list_images(self, services, tag, **kwargs):
        """List images"""
        services = services or list(self.kard.env.get_container().keys())
        if tag is None:
            tag = self.kard.meta["tag"]
        for service in services:
            write(self.make_image_name(service, tag))


class LogOutput:
    """Manage printing docker logs"""

    def __init__(self, filename=None, bufferize=False):
        """Context manager for writing to files or to stdout."""
        if filename is None:
            self.handler = sys.stdout
        else:
            self.handler = None
            self.filename = filename
        self.buffer = []
        self.bufferize = bufferize

    def __enter__(self):
        if self.handler != sys.stdout:
            self.handler = open(self.filename, "a", encoding="utf-8")
        return self

    def __exit__(self, *_):
        self.flush()
        if self.handler != sys.stdout:
            self.handler.close()
            self.handler = None

    def write(self, line):
        """Write a string to the configured output."""
        if self.bufferize:
            self.buffer.append(line)
            return
        print(line, file=self.handler, end="")
        self.handler.flush()

    def writeln(self, line):
        """Write a string followed by a newline to the configured output."""
        if self.bufferize:
            self.buffer.append(line + "\n")
            return
        print(line, file=self.handler)
        self.handler.flush()

    def write_console(self, line):
        """Write the string only when it's connected to a console."""
        if self.handler != sys.stdout:
            return
        if self.bufferize:
            self.buffer.append(line)
            return
        print(line, file=self.handler, end="")

    def flush(self):
        """Flush the handler"""
        self.handler.write("".join(self.buffer))
        self.handler.flush()



# Copyright© 1986-2024 Altair Engineering Inc.

import unittest
from pkr.utils import merge


class TestMerge(unittest.TestCase):
    """
    TestMerge tests the behavior of the merge function in pkr/utils.py.

    The current test set does not cover all possible behaviors of the merge
    function.
    """

    def test_none_source(self):
        source = None
        dest = {"key": "value"}
        result = merge(source, dest)
        self.assertEqual(result, dest)
        self.assertIsNone(source)

    def test_empty_source(self):
        source = {}
        dest = {"key": "value"}
        result = merge(source, dest)
        self.assertEqual(result, dest)

    def test_none_destination(self):
        source = {"key": "value"}
        dest = None
        result = merge(source, dest)
        self.assertEqual(result, source)
        self.assertIsNone(dest)

    def test_empty_destination(self):
        source = {"key": "value"}
        dest = {}
        result = merge(source, dest)
        self.assertEqual(result, source)

        # Mutate result, should also change dest
        result["test"] = 12
        self.assertIn("test", dest)
        self.assertEqual(dest["test"], 12)

    def test_simple_merge(self):
        source = {"key1": "value", "key2": "new value"}
        dest = {"key2": "old value", "key3": 9}
        result = merge(source, dest)

        # Result should include all unique keys and any duplicates should
        # have the value from source.
        self.assertEqual(
            result,
            {"key1": "value", "key2": "new value", "key3": 9},
        )



from .utils import pkrTestCase


class TestBaseDriver(pkrTestCase):
    PKR = "pkr"
    pkr_folder = "base_driver"
    kard_env = "dev"
    kard_driver = "base"
    kard_extra = {"flag": "flag_value"}

    def test_base_driver_values(self):
        self.kard_extra["src_path"] = self.src_path
        self.generate_kard()
        self.make_kard()

        out_dir = self.pkr_path / "kard" / "test"
        expected = sorted(
            [
                out_dir / "templated" / "folder2_dst" / "copy",
                out_dir / "templated" / "folder1" / "file2",
                out_dir / "templated" / "file1",
                out_dir / "meta.yml",
            ]
        )

        def walk(path):
            for p in path.iterdir():
                if p.is_dir():
                    yield from walk(p)
                    continue
                yield p.resolve()

        self.assertEqual(sorted(list(walk(out_dir))), expected)



# Copyright© 1986-2024 Altair Engineering Inc.

"""pkr functions for managing containers lifecycle with compose"""

from builtins import next
import sys
import os
import re
import subprocess
import time
import traceback

from pathlib import Path
import yaml

from pkr.driver import docker
from pkr.cli.log import write, debug
from pkr.utils import (
    PkrException,
    PasswordException,
    merge,
    get_current_container,
    encrypt_swap,
    decrypt_swap,
    decrypt_file,
    encrypt_with_key,
    decrypt_with_key,
)


# pylint: disable=too-few-public-methods
class ComposeConfig:
    """
    Replacement for compose.Config object.
    """

    def __init__(self, compose_config: dict):
        self.compose_config = compose_config
        self.services = [
            {"name": name, **service}
            for name, service in self.compose_config.get("services", {}).items()
        ]


# pylint: disable=no-member
class ComposePkr:
    """Implements pkr functions for docker compose"""

    COMPOSE_BIN = ["docker", "compose"]
    COMPOSE_FILE = "docker-compose.yml"
    COMPOSE_FILE_ENC = "docker-compose.enc"

    def __init__(self, kard, password=None, **kwargs):
        super().__init__(kard, password=password, **kwargs)
        self.metas["project_name"] = None
        self._base_path = None
        if self.kard is not None:
            self.compose_file = self.kard.path / self.COMPOSE_FILE
            self.compose_file_enc = self.kard.path / self.COMPOSE_FILE_ENC
            self.driver_meta = self.kard.meta.get("driver", {}).get("docker_compose", {})

    def get_meta(self, extras, kard):
        """Return the meta from the kard"""
        values = super().get_meta(extras, kard)
        # Retrieve the real_kard_path which is different if pkr run in container
        kard.meta["real_kard_path"] = str(self.get_real_kard_path())
        return values

    def _call_compose(self, *args):
        compose_cmd = [
            *self.COMPOSE_BIN,
            "-f",
            "-",
            "-p",
            self.kard.meta["project_name"],
        ] + list(args)

        debug(f"driver: _call_compose: cmd={compose_cmd}")
        compose = self._get_compose_data()
        try:
            return subprocess.run(compose_cmd, input=compose, check=True)
        except subprocess.CalledProcessError as exc:
            raise PkrException(
                f"Subcommand {compose_cmd} returned non-zero exit code: {exc.returncode}"
            ) from exc

    def _get_compose_data(self):
        if self.compose_file.exists():
            with self.compose_file.open("rb") as cp_file:
                data = cp_file.read()
        elif self.compose_file_enc.exists():
            if not self.password:
                raise PasswordException()
            data = decrypt_file(self.compose_file_enc, self.password)
        else:
            raise PkrException("Neither docker compose yaml nor encrypted file found")
        return data

    def get_templates(self):
        """Return the templates of this driver"""
        templates = super().get_templates()

        # Cleanup merged file
        if self.compose_file.exists():
            self.compose_file.unlink()

        if "compose_file" not in self.driver_meta:
            write("Warning: No docker-compose file is provided with this environment.")
            return templates

        for file in [self.driver_meta["compose_file"]] + self.driver_meta.get(
            "compose_extension_files", []
        ):
            templates.append(
                {
                    "source": self.kard.env.pkr_path / file,
                    "origin": (self.kard.env.pkr_path / file).parent,
                    "destination": "",
                    "subfolder": "compose",
                }
            )

        return templates

    def populate_kard(self, meta_txt=True):
        """Populate context for compose"""
        if "compose_file" not in self.driver_meta:
            return
        merged_compose = {}
        compose_path = self.kard.path / "compose"
        for file in compose_path.iterdir():
            # Merge the compose_file
            yaml_data = yaml.safe_load(file.open("r"))
            if yaml_data is None:
                write(f"Warning: empty template {file} is ignored\n")
            else:
                merge(yaml_data, merged_compose)

        if meta_txt:
            with self.compose_file.open("w") as dcf:
                os.chmod(self.compose_file, 0o600)
                yaml.safe_dump(merged_compose, dcf, default_flow_style=False)
        else:
            if self.password is None:
                raise PasswordException()
            yaml_str = yaml.dump(merged_compose)
            with self.compose_file_enc.open("wb") as compose_file_enc:
                compose_enc = encrypt_with_key(
                    self.password.encode("utf-8"), yaml_str.encode("utf-8")
                )
                os.chmod(self.compose_file_enc, 0o600)
                compose_file_enc.write(compose_enc)

    def _load_compose_config(self, password=None):
        if self.compose_file_enc.exists():
            pw = self.password if self.password is not None else password
            if pw is None:
                raise PasswordException()
            with self.compose_file_enc.open("rb") as compose_file_enc:
                compose_enc = compose_file_enc.read()
                compose = decrypt_with_key(pw.encode("utf-8"), compose_enc)
                compose_data = yaml.load(compose, Loader=yaml.Loader)
        else:
            with self.compose_file.open("r") as cp_file:
                compose_data = yaml.safe_load(cp_file)
        return ComposeConfig(compose_data)

    def get_real_kard_path(self):
        """Get the matching host kard path if running in a container"""
        container = get_current_container()
        kard_path = str(self.kard.path)
        if container is None:
            return self.kard.path

        # pylint: disable=unsubscriptable-object
        mount = None
        for c in container.attrs["Mounts"]:
            if kard_path.startswith(c["Destination"]):
                if mount is None or len(c["Destination"]) > len(mount["Destination"]):
                    mount = c

        if mount is not None:
            return Path(mount["Source"]) / self.kard.path.relative_to(mount["Destination"])
        return self.kard.path  # We are in container, but not a pkr one

    def _resolve_services(self, services=None):
        """Return a generator of actual services, or all if None is provided"""

        compose_config = self._load_compose_config()
        all_services = [c["name"] for c in compose_config.services]
        if services is None:
            return all_services

        # Resolve * regexp based service names
        for service in set(services):
            if "*" in service:
                reg = re.compile(service)
                services.extend([m for m in all_services if reg.match(m)])
                services.remove(service)

        # Remove unexisting services
        return set(services) & set(all_services)

    # pylint: disable=unused-argument
    def start(self, services=None, yes=False):
        """Start the stack"""
        self._call_compose("up", "-d", *(services or ()))

    # pylint: disable=too-many-locals
    def cmd_up(self, services=None, verbose=False, build_log=None):
        """Start PCLM in a docker environement.

        Use parameters stored in meta.yml to generate the
        docker-compose.yml file, and then up the container via docker-compose.

        If bare_metal support activated, configure the given host network
        interface and add an interface bridged on the host to all the
        containers that need to communicate with bare metal machines on
        the given network.
        """

        # Re-populating the context...
        self.kard.make()

        eff_modules = self._resolve_services(services)

        # This pattern is used to detect the remote image
        # pylint: disable=anomalous-backslash-in-string
        pattern = re.compile(r"^\S+\/\S+$")

        # Image names may be different from service names (e.g. image re-use)
        build_images = set(
            s["image"].partition(":")[0]  # without ":tag" suffix
            for s in self._load_compose_config().services
            if s["name"] in eff_modules and not pattern.match(s["image"])
        )

        # Revert the image name if possible
        image_pattern = self.kard.meta.get("image_pattern", self.SERVICE_VAR)
        if image_pattern:
            pattern_regex = image_pattern.replace(self.SERVICE_VAR, "(.*)")
            images_renamed = set()
            for image in build_images:
                match = re.search(pattern_regex, image)
                images_renamed.add(image if not match else match.group(1))
            build_images = images_renamed

        pull_images = []
        for s in self._load_compose_config().services:
            if s["name"] in eff_modules and pattern.match(s["image"]):
                image = s["image"].split("/")[1]
                registry = self.get_registry(
                    url=s["image"].split("/")[0], username=None, password=None
                )
                image_name = image.split(":")[0]
                remote_tag = image.split(":")[1]
                pull_images.append((image, image_name, registry, remote_tag))

        tag = self.kard.meta["tag"]

        if len(pull_images) != 0:
            self.pull_images(pull_images, tag=tag)
        if len(build_images) != 0:
            self.build_images(
                build_images, rebuild_context=False, verbose=verbose, logfile=build_log
            )

        self.start(services)

        # Do a nap while the containers are launching before calling
        # post_compose
        time.sleep(5)

        # Call post run handlers on extensions
        self.kard.extensions.post_up(eff_modules)

    def stop(self, services=None):
        """Stop the containers"""
        self._call_compose("stop", *(services or ()))

    def restart(self, services=None):
        """Restart containers"""
        self._call_compose("restart", *(services or ()))

    def get_container(self, container_name):
        """Get infos for a container"""
        containers = [
            container
            for container in self.docker.containers(filters={"name": container_name})
            if f"/{container_name}" in container["Names"]
        ]

        if len(containers) == 0:
            return None
        if len(containers) != 1:
            raise ValueError(f'ERROR: {len(containers)} containers named "{container_name}"')

        return self.docker.inspect_container(containers.pop().get("Id"))

    def get_ip(self, container):
        """Return the first IP of a container"""
        networks = container["NetworkSettings"]["Networks"]
        return next(iter(networks.values()))["IPAddress"]

    def get_status(self, container):
        """Return status of a container"""
        state = container["State"]["Status"]
        health = container["State"].get("Health", {}).get("Status")
        if health is None:
            if state == "running":
                return 0, "started"
            return 2, "stopped"
        if health == "healthy":
            return 0, "started"
        if state == "running":
            return 1, "starting"
        return 2, "stopped"

    def cmd_ps(self):
        """List containers with ips"""
        services = self._load_compose_config().services

        for service in [s["name"] for s in services]:
            container = self.get_container(self.make_container_name(service))
            if container is None:
                container_ip = "stopped"
            else:
                container_ip = self.get_ip(container)
            write(f" - {service}: {container_ip}")

    def cmd_status(self, password=None):
        """Check all containers are up and healthy"""
        services = self._load_compose_config(password).services
        status = []

        for service in services:
            service_name = service["name"]
            container = self.get_container(self.make_container_name(service_name))
            if container is None:
                if service.get("scale", 1) == 0:
                    # Ignore services that are not started by the `up` command (scale=0)
                    continue
                status.append(2)
                write(f" - {service_name}: stopped")
            else:
                container_status = self.get_status(container)
                status.append(container_status[0])
                write(f" - {service_name}: {container_status[1]}")
        if 1 in status:
            status = (1, "starting")
        elif 2 in status:
            status = (2, "down")
        else:
            status = (0, "started")
        write(f"Global status: {status[1]}")
        sys.exit(status[0])

    def clean(self, kill=False):
        """Remove the containers and the build data file.
        If -w option set, also remove all environment files
        """
        if kill:
            self._call_compose("kill")
        self._call_compose("down", "-v")

    def execute(self, container_name, *args):
        """Execute a command on a container

        Args:
          * container_name: the name of the container
          * *args: the command, like 'ps', 'aux'
        """
        execution = self.docker.exec_create(container=container_name, cmd=args, tty=True)

        return self.docker.exec_start(exec_id=execution["Id"])

    # pylint: disable=too-many-arguments
    def launch_container(self, command, image, volumes, v_1=False, links=None):
        """Generic method to launch a container"""

        if v_1:
            host_config = self.docker.create_host_config(
                binds=[":".join((v, k)) for k, v in volumes.items()],
                links={l: l for l in [self.make_container_name(s) for s in links]},
            )
            networking_config = None
        else:
            host_config = self.docker.create_host_config(
                binds=[":".join((v, k)) for k, v in volumes.items()]
            )
            network_name = self.kard.meta["project_name"] + "_default"
            networking_config = self.docker.create_networking_config(
                {network_name: self.docker.create_endpoint_config()}
            )

        container = self.docker.create_container(
            image=image,
            name=self.make_container_name("init"),
            command=command,
            host_config=host_config,
            networking_config=networking_config,
        )

        try:
            started = False
            ret = {"StatusCode": 1}
            attempt = 10
            container_id = container.get("Id")
            while started not in ("running", "exited"):
                self.docker.start(container=container_id)
                info = self.docker.inspect_container(container=container_id)
                started = info["State"]["Status"]

            while ret["StatusCode"] != 0 and attempt > 0:
                attempt -= 1
                time.sleep(3)
                ret = self.docker.wait(container=container_id)

            logs = self.docker.logs(container=container_id)
            write(logs)
        except:
            write(traceback.format_exc())
            raise
        finally:
            self.docker.remove_container(container=container_id)

        if ret["StatusCode"] != 0:
            raise PkrException(f"Container exited with non-zero status code {ret}")

    def encrypt(self, password):
        """Encrypt the compose file"""
        encrypt_swap(self.compose_file, self.compose_file_enc, password)

    def decrypt(self, password):
        """Decrypt the compose file"""
        decrypt_swap(self.compose_file, self.compose_file_enc, password)


class ComposeDriver(ComposePkr, docker.DockerDriver):
    """Combine class to create the driver"""



# Copyright© 1986-2024 Altair Engineering Inc.

"""pkr Kard"""

import copy
from getpass import getpass
import os
from pathlib import Path
import shutil

import yaml

from .driver import load_driver
from .environment import Environment
from .ext import Extensions
from .cli.log import write
from .utils import (
    PkrException,
    TemplateEngine,
    get_kard_root_path,
    merge,
    diff,
    dedup_list,
    merge_lists,
    encrypt_swap,
    decrypt_swap,
    decrypt_file,
    encrypt_with_key,
    Cmd,
)


class KardNotFound(PkrException):
    """Exception raised if the kard is not found"""


# pylint: disable=too-many-instance-attributes
class Kard:
    """Object representing the kard"""

    META_FILE = "meta.yml"
    META_FILE_ENC = "meta.enc"
    CURRENT_NAME = "current"
    LOCAL_SRC = "./src"
    CURRENT_KARD = None

    # pylint: disable=too-many-arguments
    def __init__(self, name, path, password=None, enc=Cmd.OTHER, meta=None):
        self.path = path
        self.name = name
        self.meta_file = path / self.META_FILE
        self.meta_file_enc = path / self.META_FILE_ENC
        self.password = None
        if meta is None:
            if password == "-":
                self.password = getpass()
            else:
                self.password = password

            if self.meta_file_enc.is_file():
                if enc == Cmd.ENCRYPT:
                    raise PkrException(f'Metafile for Kard "{name}" is already encrypted')
                meta_data = decrypt_file(self.meta_file_enc, self.password)
                self.clean_meta = yaml.load(meta_data, Loader=yaml.Loader)
            else:
                if enc == Cmd.DECRYPT:
                    raise PkrException(f'Metafile for Kard "{name}" is already decrypted')
                with self.meta_file.open() as meta_file:
                    self.clean_meta = yaml.safe_load(meta_file)
        else:
            self.clean_meta = meta

        if self.clean_meta["env"] is None:
            return

        self.env = Environment(
            env_name=self.clean_meta["env"], features=self.clean_meta["features"].copy()
        )

        self._compute_meta(self.clean_meta)
        self.template_meta(self.meta)

        self.real_path = Path(self.meta.get("real_kard_path", self.path))

    @property
    def extensions(self):
        """Return the Extension class loaded with the correct instance"""
        return Extensions(self.meta["features"])

    def make(self, reset=True):
        """Make the kard"""
        templates = self.driver.get_templates()

        # Reset/create all subfolders
        subfolder_list = list(map(lambda a: a.get("subfolder"), templates))
        for subfolder in [i for n, i in enumerate(subfolder_list) if i not in subfolder_list[:n]]:
            # set(map(lambda a: a.get("subfolder"), templates)):
            folder = self.path / subfolder
            if reset:
                write(f"Removing {subfolder} ... ", add_return=False)
                if folder.exists():
                    try:
                        shutil.rmtree(str(self.path / subfolder))
                    except Exception as exc:
                        write(f"Cannot delete subtree {self.path / subfolder}: {exc}")
                        raise
                    write("Done !")
                else:
                    write("Ok !")
            folder.mkdir(parents=True, exist_ok=True)

        # Copy_file / templating
        tpl_engine = self.get_template_engine()
        for template in templates:
            source = Path(self.replace_var(str(template["source"])))
            origin = Path(self.replace_var(str(template["origin"])))
            tpl_engine.copy(
                path=source,
                origin=origin,
                local_dst=self.path / template["subfolder"] / template["destination"],
                excluded_paths=[self.replace_var(e) for e in template.get("excluded_paths", [])],
                gen_template=template.get("gen_template", True),
            )

        self.extensions.populate_kard()
        self.driver.populate_kard(self.meta_file.exists())

    @classmethod
    def list(cls, kubernetes=False):
        """Return a list of all kards"""
        try:
            kards = [p.name for p in get_kard_root_path().iterdir()]
        except OSError:
            return None
        try:
            kards.pop(kards.index(cls.CURRENT_NAME))
        except ValueError:
            pass

        if kubernetes:
            kards += load_driver("k8s").list_kards()

        return sorted(kards)

    # pylint: disable=too-many-arguments,too-many-locals,too-many-branches
    @classmethod
    def create(cls, name, env, driver, extra, features=None, meta=None, password=None, **kwargs):
        """Factory method to create a new kard"""
        extras = {"features": []}
        if meta is not None:
            extras.update(yaml.safe_load(meta))
        extras.update(extra)
        for feature in dedup_list(extras["features"]):
            write(f"WARNING: Feature {feature} is duplicated in passed meta", error=True)

        try:
            extra_features = features
            if extra_features is not None:
                extra_features = extra_features.split(",")
                for feature in dedup_list(extra_features):
                    write(f"WARNING: Feature {feature} is duplicated in args", error=True)
                merge_lists(extra_features, extras["features"], insert=False)
        except AttributeError:
            pass

        # Sanitize input metas
        for key, value in list(extras.items()):
            if isinstance(value, str) and value.lower() in ("true", "false"):
                extras[key] = value = value.lower() == "true"
            if "." in key:
                extras.pop(key)
                dict_it = extras
                sub_keys = key.split(".")
                for sub_key in sub_keys[:-1]:
                    dict_it = dict_it.setdefault(sub_key, {})
                dict_it[sub_keys[-1]] = value

        # Create the folder
        get_kard_root_path().mkdir(exist_ok=True)

        kard_path = cls._build_kard_path(name)
        kard_path.mkdir(exist_ok=True)

        try:
            if driver is not None:
                extras.setdefault("driver", {})["name"] = driver
            extras["env"] = env
            # If a path is provided, we take it. Otherwise, we use a src folder
            # in the kard folder.
            src_path = extras.pop("src_path", None)
            if src_path is not None:
                extras.update({"src_path": src_path})

            kard = cls(name, kard_path, password, Cmd.OTHER, extras)

            if env is not None:
                kard.update()
        except Exception:
            # If anything happened, we remove the folder
            shutil.rmtree(str(kard_path))
            raise

        if not kwargs.get("do_not_set_current"):
            Kard.set_current(kard.name)
        return kard

    @classmethod
    def _build_kard_path(cls, name):
        return get_kard_root_path() / name

    @classmethod
    def get_current(cls):
        """Return the current kard name"""
        current_kard_path = cls._build_kard_path(cls.CURRENT_NAME)
        name = current_kard_path.resolve().name
        return name

    @classmethod
    def load(cls, name, password=None, enc=Cmd.OTHER):
        """Return a loaded context"""
        try:
            return cls(name, cls._build_kard_path(name), password, enc)
        except IOError as ioe:
            # pylint: disable=raise-missing-from
            raise KardNotFound(f'Kard "{name}" not found: {ioe}')

    @classmethod
    def load_current(cls, kard=None, password=None, enc=Cmd.OTHER):
        """Load the last used Kard, or a specific one if given."""
        if cls.CURRENT_KARD is None:
            if not kard:
                kard = os.getenv("PKR_KARD")
            if not kard:
                kard = cls.get_current()

            cls.CURRENT_KARD = cls.load(kard, password, enc)
        return cls.CURRENT_KARD

    @classmethod
    def set_current(cls, kard_name, password=None):
        """Set the current kard by making the symlink pointing to the correct
        folder.
        """
        if "/" in kard_name:
            driver, kard_name = kard_name.split("/")
            kard = cls.create(kard_name, None, driver, {}, None, None, password)
            load_driver(driver, kard, password).load_kard()
        dst_path = kard_name

        if not cls._build_kard_path(dst_path).exists():
            raise KardNotFound(f'Kard "{kard_name}" not found.')

        # Remove the link if it exists
        try:
            current_path = cls._build_kard_path(cls.CURRENT_NAME)
            current_path.unlink()
        except OSError:
            pass

        # Link the new context
        current_path.symlink_to(dst_path)
        write(f"Current kard is now: {kard_name}")

    def update(self):
        """Update the kard"""
        if self.meta_file_enc.is_file():
            yaml_str = yaml.dump(self.clean_meta)
            with self.meta_file_enc.open("wb") as meta_file_enc:
                meta_enc = encrypt_with_key(
                    self.password.encode("utf-8"), yaml_str.encode("utf-8")
                )
                meta_file_enc.write(meta_enc)
        else:
            with self.meta_file.open("w") as meta_file:
                os.chmod(self.meta_file, 0o600)
                yaml.safe_dump(self.clean_meta, meta_file, default_flow_style=False)

    def dump(self, cleaned=True, **_):
        """Dump overall templating context meta"""
        return yaml.safe_dump(self.clean_meta if cleaned else self.meta, default_flow_style=False)

    def _compute_meta(self, extra):
        """
        Compute meta

        Process:
         - Compute meta context (env + driver + kard specific)
         - Save meta for later diff
         - Call extensions `setup` which may alter meta
         - Diff what has been changed by extensions
         - Apply command line meta(s) on top
        """
        # Extract features to push them later (ensure precedence)
        features = extra.pop("features", [])

        # Copy extra to meta
        self.meta = copy.deepcopy(extra)

        # Add env to overall context in kard.meta
        merge(self.env.get_meta(extra), self.meta)  # Extra receiving ask_input values

        # Load driver an add it to overall context
        self.meta.setdefault("driver", {}).setdefault("name", "compose")  # Default value
        driver_args = self.meta.get("driver", {}).get("args", [])
        driver_kwargs = self.meta.get("driver", {}).get("kwargs", {})
        self.driver = load_driver(
            self.meta["driver"]["name"], self, self.password, *driver_args, **driver_kwargs
        )
        merge(self.driver.get_meta(extra, self), self.meta)  # Extra receiving ask_input values

        # Populate src_path before extension call
        self._process_src_path()

        # Copy overall context as diff base
        overall_context = copy.deepcopy(self.meta)

        # Extensions (give them a copy of extra, ext should not interact with it)
        self.extensions.setup(copy.deepcopy(extra), self)

        # Making a diff and apply it to extra without overwrite
        # (we want cli to superseed extensions)
        merge(diff(overall_context, self.meta), extra, overwrite=False)

        # We add all remaining extra to the meta(s) (cli superseed all)
        merge(extra, self.meta)
        self._process_src_path()

        # Append features
        merge_lists(features, self.meta["features"], insert=False)
        extra["features"] = features

        return extra

    def _process_src_path(self):
        """Set the src path in the meta"""
        if "src_path" not in self.meta:
            self.meta["src_path"] = str(self.path / self.LOCAL_SRC)
        if not Path(self.meta["src_path"]).is_absolute():
            self.meta["src_path"] = str((self.env.pkr_path / self.meta["src_path"]).resolve())

    @classmethod
    def template_meta(cls, meta, context=None):
        """Return the meta auto-templated"""
        context = context or meta
        tpl_engine = TemplateEngine(context)
        for key, value in meta.items():
            if isinstance(value, dict):
                cls.template_meta(value, context)
            elif isinstance(value, list):
                tmp = []
                for item in value:
                    if isinstance(item, str):
                        tmp.append(tpl_engine.process_string(item))
                    elif isinstance(item, dict):
                        tmp.append(cls.template_meta(item, context))
                meta[key] = tmp
            elif isinstance(value, str):
                meta[key] = tpl_engine.process_string(value)
                if meta[key].startswith("---\n"):
                    meta[key] = yaml.safe_load(meta[key])

        return meta

    def get_template_engine(self, extra_data=None):
        """Return the template engine with template data"""
        data = self.meta.copy()

        def read_kard_file(conf_file_name):
            conf_path = self.path / Path(conf_file_name).expanduser()
            return conf_path.read_text()

        def format_image(image_name):
            image = f"{image_name}:{self.meta['tag']}"
            registry = self.meta.get("registry")
            if not registry:
                return image
            return f"{registry}/{image}"

        def get_data_path(path):
            """Prefix the given path with the path to data volumes.

            Resolves a (Kard-relative or absolute) given data_path
            or goes with the default "<kard>/data"."""

            data_path = Path(self.meta.get("data_path", "data"))

            if data_path.is_absolute():
                return str(data_path / path)

            return str(self.real_path / data_path / path)

        data.update(
            {
                "env": self.env.env_name,
                "kard_file_content": read_kard_file,
                "format_image": format_image,
                "context_path": lambda p="", c=None: str(self.driver.context_path(p, c)),
                "kard_path": lambda p="": str(self.real_path / p),
                "src_path": lambda p="": str(Path(self.meta["src_path"]) / p),
                "make_container_name": self.driver.make_container_name,
                "make_image_name": lambda n, t=None: self.driver.make_image_name(n, t),
                "data_path": get_data_path,
            }
        )

        # Get custom template data from extensions
        for custom_data in self.extensions.get_context_template_data(self.password):
            if custom_data:
                data.update(custom_data)

        if extra_data:
            data.update(extra_data)

        tpl_engine = TemplateEngine(data)

        return tpl_engine

    def replace_var(self, path):
        """Replace the Kard var if present, elsif return full path"""
        kard_path_var = "$KARD_PATH"
        src_path_var = "$SRC_PATH"
        if kard_path_var in path:
            return path.replace(kard_path_var, str(self.path))
        if src_path_var in path:
            return path.replace(src_path_var, self.meta["src_path"])
        return Path(self.env.pkr_path / path)

    def encrypt(self, password):
        """Encrypt the kard"""
        encrypt_swap(self.meta_file, self.meta_file_enc, password)

    def decrypt(self, password):
        """Decrypt the kard"""
        decrypt_swap(self.meta_file, self.meta_file_enc, password)



# Copyright© 1986-2024 Altair Engineering Inc.

"""pkr functions for managing containers lifecycle with buildx"""

from concurrent.futures import ProcessPoolExecutor
import copy
import os
from pathlib import Path
import tempfile
import sys

from python_on_whales import docker, DockerException

from pkr.driver.docker import DockerDriver
from pkr.cli.log import write
from pkr.utils import merge

BUILDKIT_ENV = {
    "env.BUILDKIT_STEP_LOG_MAX_SIZE": 1000000,
    "env.BUILDKIT_STEP_LOG_MAX_SPEED": 100000000,
}
BUILDX_OPTIONS = {
    "cache_to": {
        "type": "registry",
        "mode": "max",
    },
    "cache_from": {
        "type": "registry",
    },
    "progress": "plain",
}
BUILDX_BUILDER_NAME = "pkrbuilder"


# pylint: disable=abstract-method
class BuildxDriver(DockerDriver):
    """Driver using `docker buildx` subcommands (included in docker CE 19.03 and newer) to build
    images. Inherited from docker driver (for get_templates/pull/push/download/import/list)

    Buildx provides builkit features to the build and distributed cache
    mechanism (using registry as layer storage)
    """

    DOCKER_CONTEXT = "buildx"

    def __init__(self, kard, **kwargs):
        super().__init__(kard=kard, **kwargs)
        self.metas = {"tag": None, "buildx": ["cache_registry"]}
        self.buildkit_env = BUILDKIT_ENV
        self.buildx_options = BUILDX_OPTIONS
        self.platform = os.environ.get("DOCKER_DEFAULT_PLATFORM")
        self.builder_name = None

    def get_meta(self, extras, kard):
        values = super().get_meta(extras, kard)
        if "tag" in extras:
            extras["tag"] = str(values["tag"])

        # Get values from meta if defined
        self.buildkit_env.update(kard.meta.get("buildx", {}).get("buildkit_env", {}))
        self.builder_name = kard.meta.get("buildx", {}).get("builder_name", BUILDX_BUILDER_NAME)

        # Force cache_registry to be a sub-repo (saved in metafile)
        if extras["buildx"]["cache_registry"] == "None":
            extras["buildx"]["cache_registry"] = None
        if (
            extras["buildx"]["cache_registry"] is not None
            and "/" not in extras["buildx"]["cache_registry"]
        ):
            extras["buildx"]["cache_registry"] += "/cache"

        # Compute options
        builded_options = {
            "builder": self.builder_name,
            "cache_to": {
                "ref": extras["buildx"]["cache_registry"],
            },
            "cache_from": {
                "ref": extras["buildx"]["cache_registry"],
            },
        }
        self.buildx_options = kard.meta.get("buildx", {}).get("options", self.buildx_options)

        # Handle null cache_registry
        if extras["buildx"]["cache_registry"] is None:
            del builded_options["cache_from"]
            del builded_options["cache_to"]

        merge(builded_options, self.buildx_options, overwrite=False)

        return values

    def _create_builder(self, purge=False):
        """Create the buildkit builder, never to be purged"""
        for builder in docker.buildx.list():
            if builder.name == self.builder_name:
                if purge:
                    builder.remove()
                else:
                    break
        else:
            docker.buildx.create(name=self.builder_name, driver_options=self.buildkit_env)
            write(f"Start buildx builder {self.builder_name}")
            with open("/dev/null", "a", encoding="utf-8") as devnull:
                os.dup2(sys.stdout.fileno(), 3)
                os.dup2(devnull.fileno(), sys.stdout.fileno())
                try:
                    docker.buildx.build(self.kard.path, progress=False, builder=self.builder_name)
                except DockerException:
                    pass  # Build was never intended to success, just to force builder start
                os.dup2(3, sys.stdout.fileno())

    # pylint: disable=arguments-renamed,too-many-arguments,too-many-locals
    def build_images(
        self,
        services,
        rebuild_context,
        tag=None,
        verbose=True,
        logfile=None,
        nocache=False,
        parallel=None,
        no_rebuild=False,
        clean_builder=False,
        target=None,
        **kwargs,
    ):
        """Build docker images with buildx.

        Args:
          * services: the name of the images to build
          * tag: the tag on which the image will be saved
          * verbose: verbose logs
          * logfile: separate log file for the underlying build
          * nocache: disable docker cache
          * parallel: (int|None) Number of concurrent build
          * no_rebuild: do not build if destination image exists
          * target: name of the build-stage to build in a multi-stage Dockerfile
        """
        if docker is None:
            # Handle python 3.6 here, to not impact child drivers
            raise Exception("buildx is not supported for python < 3.6")

        services = services or list(self.kard.env.get_container().keys())
        if rebuild_context:
            self.kard.make()

        buildx_meta = self.kard.meta.get("buildx", {})
        if "cache_registry_username" in buildx_meta and buildx_meta["cache_registry"] is not None:
            registry_url = buildx_meta["cache_registry"].split("/")[0]
            print(f"Logging to {registry_url}")
            docker.login(
                server=registry_url,
                username=buildx_meta.get("cache_registry_username", None),
                password=buildx_meta.get("cache_registry_password", None),
            )
        self._create_builder(purge=clean_builder)

        tag = tag or self.kard.meta["tag"]

        if parallel:
            if len(services) >= 1:
                write(f"Building docker images using {parallel} threads ...\n")
            futures = []
            with ProcessPoolExecutor(max_workers=parallel) as executor:
                for service in services:
                    futures.append(
                        executor.submit(
                            *self._build_image(
                                service,
                                tag,
                                verbose,
                                logfile,
                                nocache,
                                no_rebuild,
                                True,
                                target,
                            )
                        )
                    )
            for future in futures:
                future.result(timeout=1800)
        else:
            if len(services) >= 1:
                write("Building docker images...\n")
            for service in services:
                execution = self._build_image(
                    service,
                    tag,
                    verbose,
                    logfile,
                    nocache,
                    no_rebuild,
                    False,
                    target,
                )
                if execution is not None:
                    execution[0](*execution[1:])

    def _build_image(
        self,
        service,
        tag=None,
        verbose=True,
        logfile=None,
        nocache=False,
        no_rebuild=False,
        bufferize=None,
        target=None,
    ):
        """Build docker image.

        Args:
          * service: service to build
          * tag: the tag on which the image will be saved
          * verbose: verbose logs
          * logfile: separate log file for the underlying build
          * nocache: disable docker cache
          * no_rebuild: do not build if destination image exists
          * bufferize: keep log to print when ended
          * target: name of the build-stage to build in a multi-stage Dockerfile
        """
        image_name = self.make_image_name(service, tag)

        dockerfile = self.kard.env.get_container(service).get("dockerfile")
        if not dockerfile:
            return None
        if not target:
            target = self.kard.env.get_container(service).get("target")

        if no_rebuild:
            image = len(self.docker.images(image_name)) == 1

        if not no_rebuild or image is False:
            context = self.kard.env.get_container(service).get("context", self.DOCKER_CONTEXT)
            self.buildx_options.update(
                {
                    "context_path": str(self.kard.path / context),
                    "file": str(Path(self.kard.path / context, dockerfile)),
                    "load": True,  # load to docker repository
                    "push": False,  # push to registry
                    "tags": image_name,
                    "target": target,
                }
            )

            if self.platform is not None:
                self.buildx_options.update({"platforms": [self.platform]})

            # Handle args
            if nocache and "cache_from" in self.buildx_options:
                del self.buildx_options["cache_from"]

            return (
                self._do_build_image,
                copy.deepcopy(self.buildx_options),
                verbose,
                logfile,
                bufferize,
            )
        return None

    @staticmethod
    def _do_build_image(
        buildx_options,
        verbose=True,
        logfile=None,
        bufferize=None,
    ):
        """Method compatible with pickle to handle build
        with multiprocessing
        """
        # Handle log output
        out_file = None
        if bufferize or logfile or not verbose:
            # Replace stdout/stderr by a file (for subprocess)
            # Equivalent of exec 3>&1 4>&2 1>$(mktemp) 2>&1
            # pylint: disable=consider-using-with
            if logfile:
                out_file = open(logfile, "a", encoding="utf-8")
            else:
                out_file = tempfile.TemporaryFile()
            os.dup2(sys.stdout.fileno(), 3)
            os.dup2(sys.stderr.fileno(), 4)
            os.dup2(out_file.fileno(), sys.stdout.fileno())
            os.dup2(out_file.fileno(), sys.stderr.fileno())

        target_name = f'({buildx_options["target"]})"' if buildx_options.get("target") else ""
        write(f'Building {buildx_options["tags"]}{target_name} image...\n')

        error = None
        buffer = None
        try:
            docker.buildx.build(**buildx_options)
        except Exception as exc:
            error = exc
        finally:
            if bufferize:
                out_file.seek(0)
                buffer = out_file.read()
            if out_file is not None:
                os.dup2(3, sys.stdout.fileno())
                os.dup2(4, sys.stderr.fileno())
                out_file.close()
            if buffer is not None:
                write(buffer.decode("utf-8"))

        if error is None:
            write("done.\n")
        else:
            raise error



# Copyright© 1986-2024 Altair Engineering Inc.

"""pkr CLI parser"""

import argparse
from getpass import getpass

from pathlib import Path

from .action import ExtendAction
from .log import write
from ..driver import list_drivers
from ..ext import ExtMixin, Extensions
from ..kard import Kard
from ..utils import PkrException, create_pkr_folder, Cmd
from ..version import __version__


def get_parser():
    """Return the pkr parser"""
    pkr_parser = argparse.ArgumentParser()
    pkr_parser.set_defaults(func=lambda _: pkr_parser.print_help())
    pkr_parser.add_argument("-v", "--version", action="version", version="%(prog)s " + __version__)

    pkr_parser.add_argument("-d", "--debug", action="store_true")
    pkr_parser.add_argument("--no-env-var", action="store_true")
    pkr_parser.add_argument(
        "-p",
        "--password",
        dest="crypt_password",
        help="password to encrypt/decrypt kard metadata with or '-' to input password from "
        "terminal",
    )

    sub_p = pkr_parser.add_subparsers(title="Commands", metavar="<command>", help="<action>")

    # Stop parser
    stop_parser = sub_p.add_parser("stop", help="Stop pkr")
    add_service_argument(stop_parser)
    add_kard_argument(stop_parser)
    stop_parser.set_defaults(
        func=lambda args: Kard.load_current(args.kard, args.crypt_password).driver.stop(
            args.services
        )
    )

    # Restart parser
    restart_parser = sub_p.add_parser("restart", help="Restart pkr")
    add_service_argument(restart_parser)
    add_kard_argument(restart_parser)
    restart_parser.set_defaults(
        func=lambda args: Kard.load_current(args.kard, args.crypt_password).driver.restart(
            args.services
        )
    )

    # Start
    start_parser = sub_p.add_parser("start", help="Start pkr")
    add_service_argument(start_parser)
    add_kard_argument(start_parser)
    start_parser.add_argument("-y", "--yes", action="store_true", help="Answer yes to questions")
    start_parser.set_defaults(
        func=lambda args: Kard.load_current(args.kard, args.crypt_password).driver.start(
            args.services, args.yes
        )
    )

    # Up parser
    up_parser = sub_p.add_parser("up", help="Rebuild context, images and start pkr")
    add_service_argument(up_parser)
    up_parser.add_argument(
        "-v", "--verbose", action="store_true", help="verbose mode", default=False
    )
    up_parser.add_argument("--build-log", help="Log file for image building", default=None)
    add_kard_argument(up_parser)
    up_parser.set_defaults(
        func=lambda args: Kard.load_current(args.kard, args.crypt_password).driver.cmd_up(
            args.services, verbose=args.verbose, build_log=args.build_log
        )
    )

    # Ps parser
    parser = sub_p.add_parser("ps", help="List containers defined in the current kard")
    add_kard_argument(parser)
    parser.set_defaults(
        func=lambda args: Kard.load_current(args.kard, args.crypt_password).driver.cmd_ps()
    )

    # Status parser
    parser = sub_p.add_parser("status", help="Check all containers of the kard are healthy")
    add_kard_argument(parser)
    parser.set_defaults(
        func=lambda args: Kard.load_current(args.kard, args.crypt_password).driver.cmd_status(
            args.crypt_password
        )
    )

    # Clean parser
    parser = sub_p.add_parser("clean", help="Stop and remove containers of current kard")
    parser.add_argument("-k", "--kill", action="store_true", help="Kill (SIGKILL) before clean")
    add_kard_argument(parser, add_short_option=False)
    parser.set_defaults(
        func=lambda args: Kard.load_current(args.kard, args.crypt_password).driver.clean(args.kill)
    )

    # Kard parser
    configure_kard_parser(sub_p.add_parser("kard", help="CLI for kards manipulation"))

    # Image parser
    configure_image_parser(sub_p.add_parser("image", help="Manage docker images"))

    # List available extensions
    list_extension_parser = sub_p.add_parser("listext", help="List extensions")
    list_extension_parser.add_argument(
        "-a", "--all", action="store_true", help="Show all available extensions"
    )
    add_kard_argument(list_extension_parser)
    list_extension_parser.set_defaults(
        func=lambda args: print(
            *(
                Extensions().list()
                if args.all
                else Kard.load_current(args.kard, args.crypt_password).extensions.list()
            ),
            sep="\n",
        )
    )

    # Ext parser
    configure_ext_parser(sub_p.add_parser("ext", help="Call extension method"))

    # Init
    init_parser = sub_p.add_parser("init", help="Build a base tree structure for pkr.")
    init_parser.add_argument(
        "path", help="The path in which to init the pkr environment.", default=None
    )

    def create_env(args):
        pkr_path = Path(args.path)
        create_pkr_folder(pkr_path)
        write(f"File structure created in : {pkr_path.absolute()}")

    init_parser.set_defaults(func=create_env)

    return pkr_parser


# pylint: disable=too-many-statements
def configure_image_parser(parser):
    """Add image parser"""
    parser.set_defaults(func=lambda _: parser.print_help())
    sub_p = parser.add_subparsers(title="Commands", metavar="<command>", help="<action>")

    # Build parser
    build_parser = sub_p.add_parser("build", help="Build docker images")
    build_parser.add_argument("-t", "--tag", default=None, help="The tag for images")
    build_parser.add_argument("-T", "--target", default=None, help="The targeted stage")
    build_parser.add_argument(
        "-r", "--rebuild-context", action="store_true", default=True, help="Rebuild the context"
    )
    build_parser.add_argument(
        "-n", "--nocache", action="store_true", help="Pass nocache to docker for the build"
    )
    build_parser.add_argument(
        "-p", "--parallel", type=int, default=None, help="Number of parallel image build"
    )
    build_parser.add_argument(
        "-b", "--no-rebuild", action="store_true", help="Disable rebuild if image already exists"
    )
    build_parser.add_argument(
        "-c",
        "--clean-builder",
        action="store_true",
        help="Clean builder before build (buildx driver)",
    )
    add_service_argument(build_parser)
    add_kard_argument(build_parser)
    build_parser.set_defaults(
        func=lambda args: Kard.load_current(args.kard, args.crypt_password).driver.build_images(
            **args.__dict__
        )
    )

    # Registries parsers
    push_parser = sub_p.add_parser("push", help="Push docker images")
    login_parser = sub_p.add_parser("login", help="Login docker registry")
    pull_parser = sub_p.add_parser("pull", help="Pull docker images")

    for r in (push_parser, login_parser, pull_parser):
        add_service_argument(r)
        r.add_argument("-r", "--registry", default=None, help="The docker registry")
        r.add_argument("-u", "--username", default=None, help="The docker registry username")
        r.add_argument("-p", "--password", default=None, help="The docker registry password")

    # Push parser
    push_parser.add_argument("-t", "--tag", default=None, help="The tag for images")
    push_parser.add_argument(
        "-o", "--other-tags", default=[], nargs="+", help="Supplemental tags for images"
    )
    push_parser.add_argument(
        "--parallel", type=int, default=None, help="Number of parallel image push"
    )
    add_kard_argument(push_parser)
    push_parser.set_defaults(
        func=lambda args: Kard.load_current(args.kard, args.crypt_password).driver.push_images(
            **args.__dict__
        )
    )

    # Login parser
    login_parser.set_defaults(
        func=lambda args: Kard.load_current(
            args.kard, args.crypt_password
        ).driver.logon_remote_registry(**args.__dict__)
    )

    # Pull parser
    pull_parser.add_argument("-t", "--tag", default=None, help="The tag for images")
    pull_parser.add_argument(
        "--parallel", type=int, default=None, help="Number of parallel image pull"
    )
    pull_parser.add_argument(
        "--ignore-errors", action="store_true", help="Ignore image pull errors"
    )
    add_kard_argument(pull_parser)
    pull_parser.set_defaults(
        func=lambda args: Kard.load_current(args.kard, args.crypt_password).driver.pull_images(
            **args.__dict__
        )
    )

    # Purge parser
    purge_parser = sub_p.add_parser(
        "purge", help="Delete all images for containers of the current kard"
    )
    purge_parser.add_argument("--tag", default=None, help="Delete images with the given tag")
    purge_parser.add_argument(
        "--except-tag", default=None, help="Do not delete images with the given tag"
    )
    purge_parser.add_argument(
        "--repository", default=None, help="Delete image reference in a specified repository"
    )
    add_kard_argument(purge_parser)
    purge_parser.set_defaults(
        func=lambda args: Kard.load_current(args.kard, args.crypt_password).driver.purge_images(
            **args.__dict__
        )
    )

    # List parser
    list_parser = sub_p.add_parser(
        "list", help="List all images for containers of the current kard"
    )
    list_parser.add_argument("--tag", default=None, help="List images with the given tag")
    add_service_argument(list_parser)
    add_kard_argument(list_parser)
    list_parser.set_defaults(
        func=lambda args: Kard.load_current(args.kard, args.crypt_password).driver.list_images(
            **args.__dict__
        )
    )

    # Download parser
    download_parser = sub_p.add_parser(
        "download", help="Download all images for containers of the current kard"
    )
    download_parser.add_argument("-r", "--registry", default=None, help="The docker registry")
    download_parser.add_argument(
        "-u", "--username", default=None, help="The docker registry username"
    )
    download_parser.add_argument(
        "-p", "--password", default=None, help="The docker registry password"
    )
    download_parser.add_argument("--tag", default=None, help="Download images with the given tag")
    download_parser.add_argument(
        "--nopull", default=False, action="store_true", help="Do not pull before export"
    )
    add_service_argument(download_parser)
    add_kard_argument(download_parser)
    download_parser.set_defaults(
        func=lambda args: Kard.load_current(args.kard, args.crypt_password).driver.download_images(
            **args.__dict__
        )
    )

    # Import parser
    import_parser = sub_p.add_parser("import", help="Import all images from kard to docker")
    add_service_argument(import_parser)
    add_kard_argument(import_parser)
    import_parser.set_defaults(
        func=lambda args: Kard.load_current(args.kard, args.crypt_password).driver.import_images(
            **args.__dict__
        )
    )


# pylint: disable=too-many-locals
def configure_kard_parser(parser):
    """
    pkr - A CI tool for Cloud Manager

    Template directory:

        jinja2 template syntax can be used in template any file in
        templates/**/ directory.

        When kard is made, The jinja2 template engine will substitute the
        variable bellow:

            - driver: contains the given value to -d/--drive parameter
                      during the pkr kard make option.

            - env: contains the given value to -e/--env parameter
                   during the pkr kard make option.

            - features: a list containing given values to --features parameter
                        by default the list contains 'init'.

            - passwords: a dict containing passwords set in meta.yml

            - add_file: this function can be used inside the dockerfiles
                        templates to render them by using either the ADD
                        or VOLUME instruction

            note: All value set with --extra parameter are available in
                  template.
    """
    parser.set_defaults(func=lambda _: parser.print_help())
    sub_p = parser.add_subparsers(title="Commands", metavar="<command>", help="<action>")

    make_context = sub_p.add_parser(
        "make", help="Generate or regenerate docker-context using the current pkr " "kard"
    )
    make_context.add_argument(
        "-u",
        "--update",
        action="store_false",
        help="Update the previous docker-context, "
        "the files added in the docker-context "
        "after the previously pkr make will not "
        "be removed",
    )
    add_kard_argument(make_context)
    make_context.set_defaults(
        func=lambda args: Kard.load_current(args.kard, args.crypt_password).make(reset=args.update)
    )

    create_kard_p = sub_p.add_parser("create", help="Create a new kard")

    def _create_kard_handler(args):
        extra = {a[0]: a[1] for a in [a.split("=", 1) for a in args.__dict__.pop("extra")]}
        return Kard.create(extra=extra, **args.__dict__)

    create_kard_p.set_defaults(func=_create_kard_handler)
    create_kard_p.add_argument("name", help="The name of the kard")
    create_kard_p.add_argument("-e", "--env", default="dev", help="The environment (dev/prod)")
    create_kard_p.add_argument(
        "-d", "--driver", default=None, help=f"The pkr driver to use {list_drivers()}"
    )
    create_kard_p.add_argument(
        "-m", "--meta", type=argparse.FileType("r"), help="A file to load meta from"
    )
    create_kard_p.add_argument(
        "-f",
        "--features",
        help="Comma-separated list of features to include in the deployment. "
        "Take one or more of: elk, protractor, salt",
        default=None,
    )
    create_kard_p.add_argument(
        "--do-not-set-current",
        help='Do not change the "current" symbolic link after creating the Kard.',
        action="store_true",
    )
    create_kard_p.add_argument(
        "--extra",
        nargs="*",
        default=[],
        action=ExtendAction,
        help="Extra args",
    )

    list_kard = sub_p.add_parser("list", help="List kards")
    list_kard.add_argument(
        "-k", "--kubernetes", action="store_true", help="Query kube remote kards"
    )

    def _list_kard_handler(args):
        kards = Kard.list(args.kubernetes)
        if kards:
            write("Kards:")
            for kard in kards:
                write(f" - {kard}")
        else:
            write("No kard found.")

    list_kard.set_defaults(func=_list_kard_handler)

    get_kard = sub_p.add_parser("get", help="Get current kard")

    def _get_kard_handler(_):
        write(f"Current Kard: {Kard.get_current()}")

    get_kard.set_defaults(func=_get_kard_handler)

    dump_kard = sub_p.add_parser(
        "dump", help="Dump current kard templating context (including all values)"
    )
    dump_kard.add_argument(
        "-c",
        "--cleaned",
        action="store_true",
        help="Include only kard specific values (effectively dump the content of meta.yml)",
    )
    add_kard_argument(dump_kard)
    dump_kard.set_defaults(
        func=lambda args: write(
            Kard.load_current(args.kard, args.crypt_password).dump(**args.__dict__)
        )
    )

    load_kard = sub_p.add_parser("load", help="Load a kard")
    load_kard.set_defaults(func=lambda args: Kard.set_current(args.name, args.crypt_password))
    load_kard.add_argument("name", help="The name of the kard")

    update_kard_p = sub_p.add_parser("update", help="Update the current kard")
    add_kard_argument(update_kard_p)
    update_kard_p.set_defaults(
        func=lambda args: Kard.load_current(args.kard, args.crypt_password).update()
    )

    encrypt_kard_p = sub_p.add_parser("encrypt", help="Encrypt metadata for the current kard")
    add_kard_argument(encrypt_kard_p)

    def _encrypt_kard_handler(args):
        kard = Kard.load_current(args.kard, args.crypt_password, Cmd.ENCRYPT)
        kard.encrypt(kard.password)
        kard.driver.encrypt(kard.password)

    # pylint: disable=unnecessary-lambda
    encrypt_kard_p.set_defaults(func=lambda args: _encrypt_kard_handler(args))

    decrypt_kard_p = sub_p.add_parser("decrypt", help="Decrypt metadata for the current kard")
    add_kard_argument(decrypt_kard_p)

    def _decrypt_kard_handler(args):
        kard = Kard.load_current(args.kard, args.crypt_password, Cmd.DECRYPT)
        kard.decrypt(kard.password)
        kard.driver.decrypt(kard.password)

    # pylint: disable=unnecessary-lambda
    decrypt_kard_p.set_defaults(func=lambda args: _decrypt_kard_handler(args))


def configure_ext_parser(parser):
    """Add parser for extensions"""
    parser.set_defaults(func=lambda _: parser.print_help())
    sub_p = parser.add_subparsers(title="Extensions", metavar="<extension>", help="Extensions")

    try:
        for name, ext in Extensions.list_all().items():
            if (
                hasattr(ext, "configure_parser")
                and ext.configure_parser is not ExtMixin.configure_parser
            ):
                ext_parser = sub_p.add_parser(name, help=f"{name.capitalize()} extension features")
                ext.configure_parser(ext_parser)
    except PkrException:
        pass


def add_service_argument(parser):
    """Add parser for listing services"""
    parser.add_argument(
        "-s", "--services", nargs="+", default=None, help="List of services (default to all)"
    )


def add_kard_argument(parser, add_short_option=True):
    """Add an argument for specifying the kard"""
    options = {
        "default": None,
        "help": 'Use a different Kard than "current". You may also set PKR_KARD.',
    }
    if add_short_option:
        parser.add_argument("-k", "--kard", **options)
    else:
        parser.add_argument("--kard", **options)


def input_password(pwd):
    """Let the user input the password"""
    if pwd == "-":
        return getpass()
    return pwd



from unittest.mock import patch

from pkr.driver.docker_compose import ComposeDriver

from .utils import pkrTestCase


class TestDockerComposeDriver(pkrTestCase):
    PKR = "pkr"
    pkr_folder = "path4"
    kard_env = "dev"
    kard_driver = "compose"

    kard_extra = {"tag": "123"}

    def test_cmd_up(self):
        self.kard_extra["src_path"] = self.src_path
        self.generate_kard()

        with patch.object(ComposeDriver, "build_images", return_value=None) as mock_build:
            with patch.object(ComposeDriver, "start", return_value=None) as mock_start:
                self.up()

                mock_build.assert_called_once_with(
                    {"backend"}, rebuild_context=False, verbose=False, logfile=None
                )
                mock_start.assert_called_once()

    def test_cmd_up_with_image_pattern(self):
        self.kard_extra["src_path"] = self.src_path
        self.kard_extra["image_pattern"] = "prefix-%SERVICE%"
        self.generate_kard()

        with patch.object(ComposeDriver, "build_images", return_value=None) as mock_build:
            with patch.object(ComposeDriver, "start", return_value=None) as mock_start:
                self.up()

                mock_build.assert_called_once_with(
                    {"backend"}, rebuild_context=False, verbose=False, logfile=None
                )
                mock_start.assert_called_once()



# Copyright© 1986-2024 Altair Engineering Inc.

import os
from stat import S_IMODE
import shutil

import re
import yaml

from pkr.utils import PATH_ENV_VAR
from pkr.version import __version__
from .utils import pkrTestCase, msg_hlp


class TestCLI(pkrTestCase):
    PKR = "pkr"
    pkr_folder = "path1"
    kard_env = "dev"
    kard_driver = "compose"

    def test_help(self):
        cmd = "{} -h".format(self.PKR)

        prc = self._run_cmd(cmd)

        self.assertEqual(0, prc.returncode)

        stdout = prc.stdout.read()
        self.assertTrue(stdout.startswith(b"usage:"))

    def test_version(self):
        cmd = "{} -v".format(self.PKR)

        prc = self._run_cmd(cmd)

        self.assertEqual(0, prc.returncode)

        expected_output = "pkr {}\n".format(__version__).encode()

        stdout = prc.stdout.read()
        prc.stderr.close()
        self.assertEqual(expected_output, stdout)

    def test_kard_create_should_warn_no_pkr_path(self):
        os.environ.pop(PATH_ENV_VAR)
        cmd = "{} --debug kard create test".format(self.PKR)

        prc = self._run_cmd(cmd)
        self.assertEqual(1, prc.returncode)

        error_msg = b"Current path .* is not a valid pkr path, " b"no usable env found\n"
        stderr = prc.stderr.read()
        self.assertRegex(stderr, error_msg)

    def test_should_use_valid_pkr_path_from_env(self):
        cmd = "{} kard list".format(self.PKR)

        prc = self._run_cmd(cmd)

        self.assertEqual(0, prc.returncode, prc.stderr.read())

    def test_should_not_use_invalid_pkr_path_from_env(self):
        os.environ[PATH_ENV_VAR] = "/dev"

        cmd = "{} kard list".format(self.PKR)

        prc = self._run_cmd(cmd)
        stderr = prc.stderr.read()

        self.assertEqual(1, prc.returncode, stderr)

        error_msg = b"Given path .* is not a valid pkr path, " b"no usable env found\n"
        self.assertRegex(stderr, error_msg, stderr)

    def test_should_not_use_invalid_pkr_path_from_current_path(self):
        os.environ.pop(PATH_ENV_VAR)
        cmd = "cd {} && {} kard list".format("/", self.PKR)

        prc = self._run_cmd(cmd)
        stderr = prc.stderr.read()

        self.assertEqual(1, prc.returncode, stderr)

        error_msg = b"Current path .* is not a valid pkr path, " b"no usable env found\n"
        self.assertRegex(stderr, error_msg, stderr)

    def test_should_use_valid_pkr_path_from_current_path(self):
        cmd = "cd {} && {} kard list".format(os.environ.pop(PATH_ENV_VAR), self.PKR)

        prc = self._run_cmd(cmd)

        self.assertEqual(0, prc.returncode)

    def test_should_use_valid_pkr_path_from_current_path_child_directory(self):
        cmd = "cd {}/env/ && {} kard list".format(os.environ.pop(PATH_ENV_VAR), self.PKR)

        prc = self._run_cmd(cmd)

        self.assertEqual(0, prc.returncode)

    def test_kard_list_should_display_warn_message(self):
        cmd = "{} kard list".format(self.PKR)

        prc = self._run_cmd(cmd)

        self.assertEqual(0, prc.returncode)

        msg = b"No kard found.\n"
        stdout = prc.stdout.read()
        self.assertEqual(msg, stdout)

    def test_kard_list(self):
        self.generate_kard()

        cmd = "{} kard list".format(self.PKR)

        prc = self._run_cmd(cmd)

        self.assertEqual(0, prc.returncode)

        msg = b"Kards:\n" b" - test\n"
        stdout = prc.stdout.read()
        self.assertEqual(msg, stdout)

    def test_kard_create_with_extra(self):
        cmd = "{} kard create test --extra tag=123 --extra tag2=456".format(self.PKR)

        prc = self._run_cmd(cmd)
        stdout = prc.stdout.read()
        stderr = prc.stderr.read()

        self.assertEqual(0, prc.returncode, stdout)

        msg_err = (
            b"WARNING: Feature g is duplicated in import dev/e from env dev\n"
            b"WARNING: Feature e is duplicated in env dev\n"
            b"WARNING: Feature g is duplicated in feature e from env dev\n"
        )
        msg = b"Current kard is now: test\n"
        self.assertEqual(msg, stdout, stdout)
        self.assertEqual(msg_err, stderr, stderr)

        meta_file = self.pkr_path / "kard" / "test" / "meta.yml"
        self.assertTrue(meta_file.exists())

        expected_meta = {
            "env": "dev",
            "features": [],
            "project_name": "test",
            "tag": "123",
            "tag2": "456",
        }

        self.assertEqual(expected_meta, yaml.safe_load(meta_file.open("r")))

    def test_kard_create_with_meta(self):
        shutil.copy(str(self.src_path / "meta1.yml"), str(self.pkr_path / "meta1.yml"))

        cmd = "{} kard create test --features c,d,c --meta {}/meta1.yml".format(
            self.PKR, str(self.pkr_path)
        )

        prc = self._run_cmd(cmd)
        stdout = prc.stdout.read()
        stderr = prc.stderr.read()
        self.assertEqual(0, prc.returncode, stderr)

        msg_err = (
            b"WARNING: Feature a is duplicated in passed meta\n"
            b"WARNING: Feature c is duplicated in args\n"
            b"WARNING: Feature g is duplicated in import dev/e from env dev\n"
            b"WARNING: Feature e is duplicated in env dev\n"
            b"WARNING: Feature g is duplicated in feature e from env dev\n"
        )
        msg = b"Current kard is now: test\n"
        self.assertEqual(msg, stdout, stdout)
        self.assertEqual(msg_err, stderr, stderr)

        meta_file = self.pkr_path / "kard" / "test" / "meta.yml"
        self.assertTrue(meta_file.exists())

        expected_meta = {
            "env": "dev",
            "features": ["b", "a", "d", "c"],
            "project_name": "test",
            "tag": "123",
        }

        self.assertEqual(expected_meta, yaml.safe_load(meta_file.open("r")))

        # Test pkr kard dump
        cmd = "{} kard dump".format(self.PKR)
        prc = self._run_cmd(cmd)
        dump = yaml.safe_load(prc.stdout)
        self.assertEqual(dump.get("features", []), ["h", "g", "f", "e", "b", "a", "d", "c"])
        self.assertEqual(dump.get("env_meta"), "dummy")
        self.assertEqual(
            dump.get("src_path"), os.environ.get("RUNNER_TEMP", str(self.env_test.kard_folder))
        )
        self.assertEqual(dump.get("templated_meta"), "MTIz")
        self.assertEqual(dump.get("templated_hash").get("key"), "dHV0dQ==")
        self.assertEqual(dump.get("templated_list"), ["tutu", "dHV0dQ=="])
        self.assertEqual(dump.get("templated_inline"), ["tutu", "titi"])
        self.assertEqual(dump.get("templated_dict_list"), ["tutu", {"a": "b", "c": "dHV0dQ=="}])

    def test_kard_make(self):
        self.generate_kard()
        cmd = "{} kard make".format(self.PKR)

        prc = self._run_cmd(cmd)
        stdout = prc.stdout.read()
        stderr = prc.stderr.read()

        self.assertEqual(0, prc.returncode, msg_hlp(stdout, stderr))

        msg = b"Removing docker-context ... Ok !\n" b"Removing compose ... Ok !\n"

        self.assertEqual(msg, stdout)

    def test_image_pull_properly_fail(self):
        self.generate_kard()
        # We want to test that the command `pkr image pull` get a same error 3
        # times, because it should retry 3 times, and fail with code 1
        cmd = "{} image pull -r dummyregistry -t remote_tag".format(self.PKR)

        prc = self._run_cmd(cmd)
        stdout = prc.stdout.read()
        stderr = prc.stderr.read()
        self.assertEqual(1, prc.returncode, stdout)

        expected_cmd_output = b"Pulling backend:test from dummyregistry/backend:remote_tag..."
        expected_error_regex = b"Error while pulling the image test:"
        expected_final_print_prefix = b"ImagePullError"

        error_outputs = stdout.split(b"\n")[:-1]
        # 1 line for the command output
        # 4 lines with the same error (3 tries, and the final print)
        self.assertEqual(len(error_outputs), 4, stdout)
        self.assertEqual(error_outputs[0], expected_cmd_output)

        errors = error_outputs[1:]
        # Check for 3 lines with the same message that indicate we have
        # retried 3 times
        self.assertRegex(errors[0], expected_error_regex)
        self.assertRegex(errors[1], expected_error_regex)
        self.assertRegex(errors[2], expected_error_regex)
        # Check that the last line is the final error print with the proper
        # error type
        self.assertRegex(stderr, expected_final_print_prefix)

    def test_image_pull_properly_ignore_error(self):
        self.generate_kard()
        # We want to test that the command `pkr image pull` get an error
        # of the non-existence image, and continue the job and finally success with code 0
        cmd = "{} image pull -r dummyregistry -t remote_tag --ignore-errors".format(self.PKR)

        prc = self._run_cmd(cmd)
        stdout = prc.stdout.read()
        self.assertEqual(0, prc.returncode, stdout)

        expected_cmd_output = b"Pulling backend:test from dummyregistry/backend:remote_tag..."
        expected_error_regex = b"Error while pulling the image test:"
        expected_final_print_prefix = b"Done !"

        outputs = stdout.split(b"\n")[:-1]
        # 6 lines of output
        self.assertEqual(len(outputs), 6, stdout)
        self.assertEqual(outputs[0], expected_cmd_output)

        self.assertRegex(outputs[1], expected_error_regex)
        # Check that the last line is the final Done message print with the proper
        self.assertRegex(outputs[2], expected_final_print_prefix)

    def test_image_push(self):
        self.generate_kard()

        cmd = "{} image push -r testrepo.io -t foo -o bar".format(self.PKR)

        prc = self._run_cmd(cmd)
        stdout = prc.stdout.read()
        stderr = prc.stderr.read()
        self.assertEqual(1, prc.returncode, stdout)

        expected_cmd_output = b"Pushing backend:foo to testrepo.io/backend:foo\n"
        expected_cmd_output_2 = re.compile(
            rb"ERROR: \(ImageNotFound\) 404 Client Error for http\+docker://.*: "
            rb'Not Found \("No such image: backend:foo"\)'
        )

        error_outputs = stderr.split(b"\n")[:-1]
        # 1 line for the command output
        # 4 lines with the same error (3 tries, and the final print)
        self.assertEqual(len(error_outputs), 4, stdout)
        self.assertEqual(stdout, expected_cmd_output)
        assert re.match(expected_cmd_output_2, error_outputs[3])

    def test_explicit_kard_option(self):
        self.generate_kard()

        cmd = "{} kard create test2 --do-not-set-current --extra tag=test2".format(self.PKR)
        prc = self._run_cmd(cmd)
        stdout = prc.stdout.read()
        self.assertEqual(0, prc.returncode, stdout)

        cmd = "{} kard list".format(self.PKR)
        prc = self._run_cmd(cmd)
        self.assertEqual(0, prc.returncode)
        msg = b"Kards:\n" b" - test\n" b" - test2\n"
        stdout = prc.stdout.read()
        self.assertEqual(msg, stdout)

        cmd = "{} kard get".format(self.PKR)
        prc = self._run_cmd(cmd)
        stdout = prc.stdout.read()
        self.assertEqual(b"Current Kard: test\n", stdout)

        tag_test2 = b"tag: test2"

        # Use Kard directly through command-line argument
        cmd = "{} kard dump -k test2".format(self.PKR)
        prc = self._run_cmd(cmd)
        self.assertEqual(0, prc.returncode)
        stdout = prc.stdout.read()
        self.assertTrue(tag_test2 in stdout)

        # Use Kard directly through environment variable
        cmd = "{} kard dump".format(self.PKR)
        pkr_env = os.environ.copy()
        pkr_env["PKR_KARD"] = "test2"
        prc = self._run_cmd(cmd, env=pkr_env)
        self.assertEqual(0, prc.returncode)
        stdout = prc.stdout.read()
        self.assertTrue(tag_test2 in stdout)

        # Use default ("current") Kard.
        cmd = "{} kard dump".format(self.PKR)
        prc = self._run_cmd(cmd)
        self.assertEqual(0, prc.returncode)
        stdout = prc.stdout.read()
        self.assertTrue(tag_test2 not in stdout)

    def test_encrypt_kard(self):
        self.generate_kard()
        password = "password"

        cmd = "{} --password {} kard encrypt".format(self.PKR, password)
        prc = self._run_cmd(cmd)
        stdout = prc.stdout.read()
        stderr = prc.stderr.read()
        self.assertEqual(1, prc.returncode, msg_hlp(stdout, stderr))
        expected = re.compile(
            b"ERROR: \(FileNotFoundError\) \[Errno 2\] No such file or directory: '"
            + bytes(self.env_test.tmp_kard)
            + b"/kard/test/docker-compose.yml'"
        )
        assert re.match(expected, stderr.split(b"\n")[-2])

        # fail to encrypt again
        cmd = "{} -p {} kard encrypt".format(self.PKR, password)
        prc = self._run_cmd(cmd)
        stdout = prc.stdout.read()
        stderr = prc.stderr.read()
        self.assertEqual(1, prc.returncode, msg_hlp(stdout, stderr))
        expected = b'ERROR: (PkrException) Metafile for Kard "test" is already encrypted\n'
        self.assertEqual(stderr, expected)

        # fail to decrypt with wrong password
        cmd = "{} -p wrong kard decrypt".format(self.PKR)
        prc = self._run_cmd(cmd)
        stdout = prc.stdout.read()
        stderr = prc.stderr.read()
        self.assertEqual(1, prc.returncode, msg_hlp(stdout, stderr))
        expected = b"ERROR: (Exception) Incorrect decryption password\n"
        self.assertEqual(stderr, expected)

        # decrypt
        cmd = "{} -p {} kard decrypt".format(self.PKR, password)
        prc = self._run_cmd(cmd)
        stdout = prc.stdout.read()
        stderr = prc.stderr.read()
        self.assertEqual(1, prc.returncode, msg_hlp(stdout, stderr))
        expected = re.compile(
            rb"ERROR: \(FileNotFoundError\) \[Errno 2\] No such file or directory: '"
            + bytes(self.env_test.tmp_kard)
            + rb"/kard/test/docker-compose.enc'"
        )
        assert re.match(expected, stderr.split(b"\n")[-2])

        # succeed in some regular operations
        cmd = "{} kard dump".format(self.PKR)
        prc = self._run_cmd(cmd)
        self.assertEqual(0, prc.returncode)

        cmd = "{} kard list".format(self.PKR)
        prc = self._run_cmd(cmd)
        stdout = prc.stdout.read()
        stderr = prc.stderr.read()
        self.assertEqual(0, prc.returncode, msg_hlp(stdout, stderr))

        # fail to decrypt again
        cmd = "{} --password {} kard decrypt".format(self.PKR, password)
        prc = self._run_cmd(cmd)
        stdout = prc.stdout.read()
        stderr = prc.stderr.read()
        self.assertEqual(1, prc.returncode, msg_hlp(stdout, stderr))
        expected = b'ERROR: (PkrException) Metafile for Kard "test" is already decrypted'
        self.assertEqual(stderr.split(b"\n")[-2], expected)


class TestCLIProd(pkrTestCase):
    PKR = "pkr"
    pkr_folder = "path2"
    kard_env = "prod"

    def test_kard_driver_coming_from_env(self):
        self.generate_kard()
        self.make_kard()

        k8s_path = self.pkr_path / "kard" / "test" / "k8s"
        self.assertTrue(k8s_path.exists())

    def test_src_path_coming_from_env(self):
        self.generate_kard()
        self.make_kard()


class TestKardMake(pkrTestCase):
    kard_extra = {"tag": "test", "foo": "bar"}
    pkr_folder = "path1"
    kard_env = "dev"
    kard_driver = "compose"

    def test_docker_compose_present(self):
        self.generate_kard()
        self.make_kard()

        # Check if files are created
        dc_file = self.pkr_path / "kard" / "current" / "docker-compose.yml"
        self.assertTrue(dc_file.exists())

        expected_docker_compose = {
            "services": {},
            "paths": [
                str(self.context_path.resolve() / "test"),
                str(self.env_test.kard_folder / "test"),
            ],
        }

        self.assertEqual(expected_docker_compose, yaml.safe_load(dc_file.open("r")))

    def test_dockerfile_present(self):
        self.generate_kard()
        self.make_kard()

        gen_file = self.context_path / "backend.dockerfile"
        expected_content = """FROM python:alpine3.7\n"""

        self.assertTrue(gen_file.exists())

        content = gen_file.open("r").read()
        self.assertEqual(expected_content, content)

    def test_config_present(self):
        self.generate_kard()
        self.make_kard()

        gen_file = self.context_path / "backend" / "backend.conf"
        expected_content = (
            "foo=bar\n"
            "tag_b64=dGVzdA==\n"
            "tag_sha256=9f86d081884c7d659a2feaa0c55ad015a3bf4f1b2b0b822cd15d6c15b0f00a08\n"
            "htpassword=a:"
        )

        self.assertTrue(gen_file.exists())

        content = gen_file.open("r").read()
        self.assertEqual(expected_content, content)


class TestKardMakeWithRelativeSrcPath(pkrTestCase):
    kard_extra = {"tag": "test", "src_path": "dummy/.."}
    pkr_folder = "path1"
    kard_env = "dev"
    # test_clean = False

    def test_docker_compose_present(self):
        self.generate_kard()
        self.make_kard()

        # Check if files are created
        dc_file = self.pkr_path / "kard" / "current" / "docker-compose.yml"
        self.assertTrue(dc_file.exists())

        expected_docker_compose = {
            "services": {},
            "paths": [
                str(self.context_path.resolve() / "test"),
                str(self.pkr_path / "test"),
            ],
        }

        self.assertEqual(expected_docker_compose, yaml.safe_load(dc_file.open("r")))

    def test_start_script_present_with_proper_mod(self):
        self.generate_kard()
        self.make_kard()

        gen_file = self.context_path / "backend" / "start.sh"
        expected_content = "#!/bin/bash\n\nsleep 1\n"

        self.assertTrue(gen_file.exists())

        content = gen_file.open("r").read()
        self.assertEqual(expected_content, content)

        self.assertTrue(os.access(gen_file, os.R_OK))
        self.assertTrue(os.access(gen_file, os.X_OK))

    def test_exec_script_present_with_proper_mod(self):
        self.generate_kard()
        self.make_kard()

        gen_file = self.context_path / "backend" / "exec.sh"
        expected_content = "echo test"

        self.assertTrue(gen_file.exists())

        content = gen_file.open("r").read()
        self.assertEqual(expected_content, content)

        self.assertTrue(os.access(gen_file, os.R_OK))
        self.assertTrue(os.access(gen_file, os.X_OK))


class TestKardMakeWithExtensionDev(TestKardMake):
    pkr_folder = "path2"
    kard_env = "dev"

    def test_docker_compose_present(self):
        self.generate_kard()
        self.make_kard()

        # Check if files are created
        dc_file = self.pkr_path / "kard" / "current" / "docker-compose.yml"
        self.assertTrue(dc_file.exists())

        expected_docker_compose = {
            "services": {},
            "paths": [
                str(self.context_path.resolve() / "test"),
                str(self.pkr_path / "kard" / "test" / "src" / "test"),
            ],
        }

        self.assertEqual(expected_docker_compose, yaml.safe_load(dc_file.open("r")))

    def test_dockerfile_present(self):
        self.generate_kard()
        self.make_kard()

        gen_file = self.context_path / "backend.dockerfile"
        expected_content = "FROM python:alpine3.7\n\n" 'VOLUME ["/usr/src/app"]'

        self.assertTrue(gen_file.exists())

        content = gen_file.open("r").read()
        self.assertEqual(expected_content, content)


class TestKardMakeWithExtensionProd(TestKardMake):
    pkr_folder = "path2"
    kard_env = "prod"

    def test_docker_compose_present(self):
        self.generate_kard()
        self.make_kard()

        # Check if files are created
        dc_file = self.pkr_path / "kard" / "current" / "docker-compose.yml"
        self.assertTrue(dc_file.exists())

        expected_docker_compose = {
            "services": {},
            "paths": [
                str(self.context_path.resolve() / "test"),
                str(self.pkr_path / "kard" / "test" / "src" / "test"),
            ],
        }

        self.assertEqual(expected_docker_compose, yaml.safe_load(dc_file.open("r")))

    def test_dockerfile_present(self):
        self.generate_kard()
        self.make_kard()

        gen_file = self.context_path / "backend.dockerfile"
        expected_content = "FROM python:alpine3.7\n\n" 'ADD "app" "/usr/src/app"'

        self.assertTrue(gen_file.exists())

        content = gen_file.open("r").read()
        self.assertEqual(expected_content, content)



# Copyright© 1986-2024 Altair Engineering Inc.

"""Module with the pkr environment"""

import yaml

from .cli.log import write
from .utils import (
    ENV_FOLDER,
    HashableDict,
    ensure_definition_matches,
    get_pkr_path,
    merge,
    dedup_list,
    merge_lists,
)


class Environment:
    """Class for loading and holding pkr environment"""

    IMPORT_KEY = "import"
    DEFAULT_TEMPLATE_DIR = "templates"

    def __init__(self, env_name, features=None, path=None):
        self.env_name = env_name
        self.pkr_path = get_pkr_path()
        self.path = path or self.default_path

        # First load the main file to add eventual dependencies
        env_path = self.path / env_name
        env_file_path = env_path / "env.yml"

        self.env = self._load_env_file(env_file_path)

        self.features = features or []
        for feature in dedup_list(self.env.get("default_features", [])):
            write(f"WARNING: Feature {feature} is duplicated in env {env_name}", error=True)
        merge_lists(self.env.get("default_features", []), self.features)

        for feature in self.features:
            f_path = env_path / (feature + ".yml")
            if f_path.is_file():
                content = self._load_env_file(f_path)
                feature_features = content.pop("default_features", [])
                for dup in dedup_list(feature_features):
                    write(
                        f"WARNING: Feature {dup} is duplicated in feature {feature} from env "
                        f"{env_name}",
                        error=True,
                    )
                merge_lists(feature_features, self.features)
                content["default_features"] = feature_features
                merge(content, self.env)

    def _load_env_file(self, path):
        """Load an environment with its dependencies recursively"""
        with path.open() as env_file:
            content = yaml.safe_load(env_file)
            if content is None:
                content = {}
            if "default_features" not in content:
                content["default_features"] = []

        for imp_name in content.get(self.IMPORT_KEY, ()):
            imp_path = self.path / (imp_name + ".yml")
            imp_data = self._load_env_file(imp_path)
            imp_data.pop(self.IMPORT_KEY, None)
            imp_features = imp_data.pop("default_features", [])
            for dup in dedup_list(imp_features):
                write(
                    f"WARNING: Feature {dup} is duplicated in import {imp_name} from env "
                    f"{self.env_name}",
                    error=True,
                )
            merge_lists(imp_features, content["default_features"])
            content = merge(content, imp_data)
        return content

    def get_meta(self, extra):
        """Ensure that all metadata are present"""
        default = self.env.get("default_meta")

        if "driver" in self.env:
            write(
                f"DEPRECATION: driver should be under default_meta in env {self.env_name}",
                error=True,
            )
            merge(self.env["driver"], default.setdefault("driver", {}), overwrite=False)

        if not default:  # This prevent an empty value in the YAML
            default = {}

        ret = default.copy()
        merge(extra, ret)

        required_values = ensure_definition_matches(
            definition=self.env.get("required_meta", []), defaults=ret, data=extra
        )
        merge(required_values, extra)

        # Feature
        ret["features"] = self.env.get("default_features", [])
        return ret

    @property
    def default_path(self):
        """Return the default path"""
        return self.pkr_path / ENV_FOLDER

    def _containers(self, template=False):
        """Method for fetching the containers dict as the schema might
        evolve.

        Args:
          - template: a bool specifying if templates should be returned
        """

        containers = self.env["containers"]

        if template:
            return containers

        if not containers:
            return {}

        return {
            name: value
            for name, value in containers.items()
            if value and not value.get("template", False)
        }

    @property
    def context_dir(self):
        """Return the context folder name"""
        return self.env["context_dir"]

    @property
    def template_dir(self):
        """Return the template folder name"""
        return self.env.get("template_dir", self.DEFAULT_TEMPLATE_DIR)

    def get_container(self, name=None):
        """Return a compiled dictionary representing a container, or a list of
        all if name is not specified.

        Args:
          - name: the name of the container to retrieve
        """
        if name is None:
            ret = {}
            for c_name, cont in self._containers().items():
                if not cont.get("template", False):
                    ret[c_name] = self.get_container(c_name)
            return ret

        container = self._containers(template=True)[name] or {}
        if "parent" in container and container["parent"] is not None:
            parent = self.get_container(container["parent"])
            return merge(container, parent.copy())

        return container

    def get_requires(self, containers=None):
        """Returns a list of required files for the provided containers.

        The result is returned as a list of dicts with 3 values: origin, src
        and dst.

        Args:
          - containers: a list of containers name
        """

        if containers is None:
            containers = list(self._containers().keys())

        requirements = {}
        # We first put them in a dict containing sets to avoid having doubles
        for name in containers:
            env = self.get_container(name)
            for key, value in env.get("requires", {}).items():
                dst_set = requirements.get(key, set())
                dst_set.add(HashableDict(value))
                requirements[key] = dst_set

        # then we transform it to a list of dicts
        ret = []
        for key, values in requirements.items():
            for value in values:
                item = {"origin": key}
                item.update(value)
                ret.append(item)

        return ret

    def __getitem__(self, item):
        return self.env.__getitem__(item)

    # pylint: disable=unnecessary-dunder-call
    def get(self, item, default):
        """Return value from the env"""
        try:
            return self.__getitem__(item)
        except KeyError:
            return default



# Copyright© 1986-2024 Altair Engineering Inc.

"""Pkr functions for handling the life cycle with k8s"""
import base64
import os
import shlex
import subprocess
import sys
import zlib
from tempfile import NamedTemporaryFile
from time import sleep

from kubernetes import client, config
import yaml

from pkr.driver import docker
from pkr.cli.log import write

CONFIGMAP = {
    "apiVersion": "v1",
    "kind": "ConfigMap",
    "metadata": {
        "namespace": "kube-system",
        "labels": {
            "pkr": "kard",
        },
    },
    "data": {},
}


class KubernetesPkr:
    """K8s implementation"""

    K8S_FOLDER = "k8s"
    K8S_CONFIG = os.path.expandvars("$KUBECONFIG")

    # pylint: disable=no-member
    def __init__(self, kard, password=None, **kwargs):
        super().__init__(kard, password=password, **kwargs)

        self.metas["registry"] = None
        self._client = None
        if hasattr(self.kard, "meta"):
            self.namespace = self.kard.meta.get("k8s", {}).get("namespace", "default")

        self.env = {
            "KUBECONFIG": self.K8S_CONFIG,
            "PATH": os.environ.get("PATH"),
        }

    @property
    def client(self):
        """Instantite a client, cache it and returns it"""
        if not self._client:
            config.load_kube_config(self.K8S_CONFIG)
            self._client = client.CoreV1Api()
        return self._client

    def get_templates(self):
        """Return the k8s templates"""
        templates = super().get_templates()

        for file in self.kard.meta["driver"].get("k8s", {}).get("k8s_files", []):
            templates.append(
                {
                    "source": self.kard.env.pkr_path / file,
                    "origin": (self.kard.env.pkr_path / file).parent,
                    "destination": "",
                    "subfolder": "k8s",
                }
            )

        return templates

    def run_cmd(self, command, silent=False):
        """Run a shell command"""
        xargs = {}
        if silent:
            xargs["stdout"] = subprocess.PIPE
            xargs["stderr"] = subprocess.PIPE
        with subprocess.Popen(shlex.split(command), env=self.env, close_fds=True, **xargs) as proc:
            stdout, stderr = proc.communicate()

        return stdout or "", stderr or "", proc.returncode

    def run_kubectl(self, cmd, silent=False):
        """Run kubectl tool with the provided command"""
        return self.run_cmd(f"kubectl {cmd}", silent)

    def new_configmap(self):
        """
        Provide a fresh configmap
        """
        cm = CONFIGMAP.copy()
        cm["metadata"]["name"] = f"pkr-{self.kard.name}"
        return cm

    def get_configmap(self):
        """
        Retrieve previously stored content for this kard
        """
        out, err, returncode = self.run_kubectl(
            f"get cm -n kube-system pkr-{self.kard.name} -o yaml", silent=True
        )
        if returncode and err != "":
            raise Exception(f"Failed to get configmap pkr-{self.kard.name} with : {err}")
        out_hash = {}
        for key, value in yaml.safe_load(out).get("data", {}).items():
            out_hash[key] = zlib.decompress(base64.b64decode(value)).decode("utf-8")
        return out_hash

    def write_configmap(self, c_m):
        """
        Write a configmap for this kard with deployed content
        cm: {"pkr_template_name": "pkr_template_content", ...}
        """
        if len(c_m) == 0:
            self.run_kubectl(f"delete cm -n kube-system pkr-{self.kard.name}")
            return
        cm_compressed = self.new_configmap()
        for key in c_m:
            cm_compressed["data"][key] = base64.b64encode(zlib.compress(c_m[key].encode("utf-8")))
        with NamedTemporaryFile() as f:
            f.write(yaml.dump(cm_compressed).encode("utf-8"))
            self.run_kubectl(f"apply -f {f.name}")

    # pylint: disable=too-many-branches
    def start(self, services=None, yes=False):
        """Starts services

        Args:
          * services: a list with the services name to start
        """
        k8s_files_path = self.kard.path / "k8s"
        meta_file = self.kard.path / "meta.yml"
        saved_files = [meta_file] + sorted(k8s_files_path.glob("*.yml"))

        old_cm = self.get_configmap()
        new_cm = {}

        write("Compare with previous deployment ...")
        for k8s_file in saved_files:
            if services and k8s_file.name[:-4] not in services:
                new_cm[k8s_file.name] = old_cm[k8s_file.name]
                continue

            if k8s_file.name in old_cm:
                with NamedTemporaryFile() as f:
                    f.write(old_cm[k8s_file.name].encode("utf-8"))
                    f.seek(0)
                    self.run_cmd(f"diff -u {f.name} {k8s_file}")
            else:
                if not k8s_file.name.startswith("."):
                    write(f"Added file {k8s_file}")

            if not k8s_file.name.startswith("."):
                with open(str(k8s_file), "r", encoding="utf-8") as f:
                    new_cm[k8s_file.name] = f.read()

        for name in old_cm:
            if name not in new_cm:
                write(f"Removed file {name}")

        if sys.stdin.isatty() and not yes:
            proceed = None
            while proceed not in {"y", "n"}:
                proceed = input("Apply (y/n) ? ")
            if proceed == "n":
                return

        write("\nApplying manifests ...")
        for name, cm_content in old_cm.items():
            if name not in new_cm:
                write(f"Removing {name}")
                with NamedTemporaryFile() as f:
                    f.write(cm_content.encode("utf-8"))
                    f.seek(0)
                    out, _, _ = self.run_kubectl(f"delete -f {f.name}")
                    write(out)

        for k8s_file in saved_files[1:]:
            if services and k8s_file.name[:-4] not in services:
                continue
            write(f"Processing {k8s_file.name}")
            out, _, _ = self.run_kubectl(f"apply -f {k8s_file}")
            write(out)
            sleep(0.1)

        self.write_configmap(new_cm)

    def stop(self, services=None):
        """Stops services"""
        k8s_files_path = self.kard.path / "k8s"

        for k8s_file in sorted(k8s_files_path.glob("*.yml"), reverse=True):
            if services and k8s_file.name[:-4] not in services:
                continue
            write(f"Processing {k8s_file}")
            out, _, _ = self.run_kubectl(f"delete -f {k8s_file}")
            write(out)
            sleep(0.5)

        self.write_configmap({})

    def restart(self, services=None):
        """Restart services"""
        raise NotImplementedError()

    def cmd_ps(self):
        """List containers with ips"""
        response = self.client.list_namespaced_pod(self.namespace)
        services = response.items
        for service in services:
            write(f" - {service.metadata.name}: {service.status.phase} - {service.status.pod_ip}")

    # pylint: disable=unused-argument
    def clean(self, kill=False):
        """Clean the stack"""
        self.stop()

    def list_kards(self):
        """Return the list of kards stored in the config map"""
        out, err, returncode = self.run_kubectl("get cm -n kube-system -l pkr=kard -o name", True)
        if returncode and err != "":
            raise Exception(f"Error getting kards from k8s: {err}")

        kards = []
        if out != "":
            for line in out.decode("utf-8").splitlines():
                _, name = line.split("configmap/pkr-")
                kards.append(f"k8s/{name}")
        return kards

    def load_kard(self):
        """Load kard from config map"""
        cm = self.get_configmap()
        with self.kard.meta_file.open("w+") as meta_file:
            meta_file.write(cm["meta.yml"])

    def encrypt(self, password=None):
        """Encrypt the kard"""
        raise NotImplementedError()

    def decrypt(self, password=None):
        """Decrypt the kard"""
        raise NotImplementedError()


# pylint: disable=abstract-method
class KubernetesDriver(KubernetesPkr, docker.DockerDriver):
    """Combine classes for creating the driver"""



import unittest
import yaml
from pkr import driver
from .utils import pkrTestCase


class TestDriver(unittest.TestCase):
    def test_list_drivers(self):
        self.assertEqual(
            driver.list_drivers(),
            ("base", "buildx", "buildx_compose", "buildx_k8s", "docker", "docker_compose", "k8s"),
        )

    def test_load_driver(self):
        self.assertTrue(
            repr(driver.load_driver("compose")).startswith(
                "<pkr.driver.docker_compose.ComposeDriver object"
            )
        )

    def test_load_driver_by_alias(self):
        self.assertTrue(
            repr(driver.load_driver("compose")).startswith(
                "<pkr.driver.docker_compose.ComposeDriver object"
            )
        )


class TestCompose(pkrTestCase):
    pkr_folder = "path3"
    kard_env = "dev"

    def test_expand_var(self):
        # Copy pkr_path (from requires)
        self.kard_extra["src_path"] = self.pkr_path
        self.generate_kard()
        self.make_kard()

        out_dir = self.pkr_path / "kard" / "test"
        self.assertTrue((out_dir / "docker-compose.yml").exists())

        # Check replacement of $SRC_PATH and $KARD_PATH in requires
        self.assertTrue((out_dir / "docker-context" / "copy" / "env").exists())
        self.assertFalse((out_dir / "docker-context" / "copy" / "kard" / "test").exists())
        self.assertTrue((out_dir / "docker-context" / "copied_meta.yml").exists())

        expected = [
            str(self.pkr_path / "kard" / "test" / "docker-context" / "test"),
            str(self.pkr_path / "test"),
            str(self.pkr_path / "kard" / "test" / "test"),
            str(self.pkr_path / "kard" / "test" / "data" / "test"),
            "container-test",
            "image-test",
        ]
        with open((out_dir / "docker-compose.yml"), "r") as f:
            dump = yaml.load(f, Loader=yaml.FullLoader)

        self.assertEqual(dump["paths"], expected)



import sys
from .utils import pkrTestCase


class TestBuildxDriver(pkrTestCase):
    PKR = "pkr"
    pkr_folder = "docker_driver"
    kard_env = "dev"
    kard_driver = "buildx_compose"
    kard_extra = {
        "tag": "123",
        "flag": "flag_value",
        "buildx.cache_registry": "dummy",
        "buildx.builder_name": "testpkrbuilder",
    }

    def test_docker_driver_values(self):
        self.kard_extra["src_path"] = self.src_path
        self.generate_kard()
        self.make_kard()

        out_dir = self.pkr_path / "kard" / "test"
        expected = sorted(
            [
                out_dir / "buildx" / "folder2_dst" / "copy",
                out_dir / "buildx" / "file1" / "file2",
                out_dir / "buildx" / "file1.dockerfile",
                out_dir / "meta.yml",
            ]
        )

        def walk(path):
            for p in path.iterdir():
                if p.is_dir():
                    yield from walk(p)
                    continue
                yield p.resolve()

        self.assertEqual(sorted(list(walk(out_dir))), expected)

    def test_docker_multiple_contexts(self):
        self.kard_extra["src_path"] = self.src_path
        self.generate_kard(env="contexts")
        self.make_kard()

        out_dir = self.pkr_path / "kard" / "test"
        self.assertTrue((out_dir / "buildx" / "folder2_dst" / "copy").exists())
        self.assertTrue((out_dir / "context1" / "folder2_dst" / "copy").exists())

        self.assertTrue((out_dir / "buildx" / "file1.dockerfile").exists())
        self.assertTrue((out_dir / "context1" / "file1.dockerfile").exists())

        cmd = "{} image build -s container1 -c".format(self.PKR)
        prc = self._run_cmd(cmd)
        stdout = prc.stdout.read()
        stderr = prc.stderr.read()

        # Python 3.6
        if sys.version_info < (3, 7):
            self.assertEqual(stdout, b"", stdout)
            self.assertEqual(
                stderr, b"ERROR: (Exception) buildx is not supported for python < 3.6\n"
            )
            return

        expected = (
            b"Warning: No docker-compose file is provided with this environment.\n"
            b"Removing context1 ... Done !\n"
            b"Removing buildx ... Done !\n"
            b"Start buildx builder testpkrbuilder\n"
            b"Building docker images...\n\n"
            b"Building container1:123 image...\n\n"
        )
        self.assertTrue("unknown instruction: flag_value" in stderr.decode("utf-8"), stderr)
        self.assertEqual(stdout, expected, stdout)
        self.assertRegex(
            stderr.decode("utf-8"),
            r"docker buildx build --progress plain --builder testpkrbuilder --load "
            r"--file " + str(self.env_test.kard_folder) + "/.*/file1.dockerfile --cache-from "
            r"type=registry,ref=dummy/cache --cache-to type=registry,mode=max,ref=dummy/cache "
            r"--tag container1:123 " + str(self.env_test.kard_folder) + r"/.*/context1",
        )



# Copyright© 1986-2024 Altair Engineering Inc.

"""pkr functions for managing containers lifecycle with buildx and k8s"""

from .k8s import KubernetesPkr
from .buildx import BuildxDriver


# pylint: disable=abstract-method
class BuildxComposeDriver(KubernetesPkr, BuildxDriver):
    """Driver using `docker buildx` subcommands with k8s"""

    def get_templates(self):
        return KubernetesPkr.get_templates(self)

    def build_images(self, *args, **kwargs):
        return BuildxDriver.build_images(self, *args, **kwargs)



# Copyright© 1986-2024 Altair Engineering Inc.

"""pkr drivers"""

from importlib import import_module
import pkgutil
import os

from .base import AbstractDriver

_USE_ENV_VAR = True

DRIVER_MAPPING = {
    "compose": "docker_compose",
    "kubernetes": "k8s",
}


def set_use_env_var(use_env_var=True):
    """Activate the usage of environment variable for Docker"""
    # pylint: disable=global-statement
    global _USE_ENV_VAR
    _USE_ENV_VAR = use_env_var


def _get_driver_class(module):
    for attr in dir(module):
        ext_cls = getattr(module, attr)
        try:
            if issubclass(ext_cls, AbstractDriver) and ext_cls is not AbstractDriver:
                return ext_cls
        except TypeError:
            pass
    return None


def load_driver(driver_name, kard=None, password=None, **kwargs):
    """Return the loaded driver"""
    driver_name = DRIVER_MAPPING.get(driver_name, driver_name)
    module = import_module(f"pkr.driver.{driver_name}", "pkr.driver")
    return _get_driver_class(module)(kard, password, **kwargs)


def list_drivers() -> tuple:
    """Return a list of drivers"""
    drivers_dir = os.path.dirname(os.path.realpath(__file__))
    return tuple(package_name for _, package_name, _ in pkgutil.iter_modules([drivers_dir]))



from .utils import pkrTestCase
import re


class TestDockerDriver(pkrTestCase):
    PKR = "pkr"
    pkr_folder = "docker_driver"
    kard_env = "dev"
    kard_driver = "docker"
    kard_extra = {"tag": "123", "flag": "flag_value"}

    def test_docker_driver_values(self):
        self.kard_extra["src_path"] = self.src_path
        self.generate_kard()
        self.make_kard()

        out_dir = self.pkr_path / "kard" / "test"
        expected = sorted(
            [
                out_dir / "docker-context" / "folder2_dst" / "copy",
                out_dir / "docker-context" / "file1" / "file2",
                out_dir / "docker-context" / "file1.dockerfile",
                out_dir / "meta.yml",
            ]
        )

        def walk(path):
            for p in path.iterdir():
                if p.is_dir():
                    yield from walk(p)
                    continue
                yield p.resolve()

        self.assertEqual(sorted(list(walk(out_dir))), expected)

    def test_docker_multiple_contexts(self):
        self.kard_extra["src_path"] = self.src_path
        self.generate_kard(env="contexts")
        self.make_kard()

        out_dir = self.pkr_path / "kard" / "test"
        self.assertTrue((out_dir / "docker-context" / "folder2_dst" / "copy").exists())
        self.assertTrue((out_dir / "context1" / "folder2_dst" / "copy").exists())

        self.assertTrue((out_dir / "docker-context" / "file1.dockerfile").exists())
        self.assertTrue((out_dir / "context1" / "file1.dockerfile").exists())

        cmd = "{} image build -s container1".format(self.PKR)
        prc = self._run_cmd(cmd)
        stdout = prc.stdout.read()
        stderr = prc.stderr.read()

        expected = (
            b"Removing context1 ... Done !\n"
            b"Removing docker-context ... Done !\n"
            b"Building container1:123 image...\n"
        )
        self.assertEqual(stdout, expected)
        expected = re.compile(b".*: unknown instruction: [Ff][Ll][Aa][Gg]_[Vv][Aa][Ll][Uu][Ee]")
        assert re.match(expected, stderr)



# Copyright© 1986-2024 Altair Engineering Inc.

"""Base definition for drivers"""

from abc import ABC
import re

from ..utils import (
    merge,
    ensure_definition_matches,
)


# pylint: disable=missing-class-docstring,too-many-public-methods
class AbstractDriver:
    """Abstract class for a driver"""

    SERVICE_VAR = "%SERVICE%"

    def __init__(self, kard, password=None, **_):
        """Save the kard to driver

        args and kwargs passed to driver
        """
        self.kard = kard
        self.metas = {}
        self.password = password

    def get_meta(self, extras, kard):
        """Ensure that the required meta are present.

        Args:
          * extras(dict): extra values
          * kard: the current kard
        """
        default = kard.env.get("default_meta", {}).copy()
        default.setdefault("project_name", re.sub(r"[^-_a-z0-9]", "", str(kard.path.name).lower()))
        merge(extras, default)

        values = ensure_definition_matches(definition=self.metas, defaults=default, data=kard.meta)
        merge(values, extras)
        return values

    def get_templates(self):
        """Return files or folders to be templated by pkr

        Format:
        [
            {
                "source": full_path,
                "origin": base_template_path (use to get relative path),
                "destination": relative_path,
                "subfolder": name of first level folder (driver specific),
                "excluded_paths": exclude those paths if source is a dir,
            }
            ...
        ]
        """
        raise NotImplementedError()

    def populate_kard(self, meta_txt=True):
        """Driver method executed after templating

        Might create some more files in kard_path, but must take care of cleaning it.
        """

    def context_path(self, sub_path, container):
        """AbstractDriver hook for driver to compute and return path
        relative to container name
        """
        raise NotImplementedError()

    #
    # Image related functions
    #
    def make_image_name(self, service, tag=None):
        """Return the image name formatted with the pattern in metas."""
        image_pattern = self.kard.meta.get("image_pattern", self.SERVICE_VAR)
        image_name = image_pattern.replace(self.SERVICE_VAR, service)
        if tag is not None:
            image_name = ":".join((image_name, tag))
        return image_name

    def build_images(self, *args, **kwargs):
        """Build images"""
        raise NotImplementedError()

    def push_images(self, *args, **kwargs):
        """Push images"""
        raise NotImplementedError()

    def download_images(self, *args, **kwargs):
        """Download images"""
        raise NotImplementedError()

    def import_images(self, *args, **kwargs):
        """Import images"""
        raise NotImplementedError()

    def list_images(self, *args, **kwargs):
        """List images"""
        raise NotImplementedError()

    def purge_images(self, *args, **kwargs):
        """Purge images"""
        raise NotImplementedError()

    #
    # Deploy related functions
    #
    def make_container_name(self, name):
        """Return the container name formatted with the pattern in metas."""
        container_pattern = self.kard.meta.get("container_pattern", self.SERVICE_VAR)
        return container_pattern.replace(self.SERVICE_VAR, name)

    def start(self, services, yes):
        """Starts services

        Args:
          * services: a list with the services name to start
        """
        raise NotImplementedError()

    def stop(self, services=None):
        """Starts services

        Args:
          * services: a list with the services name to start
        """
        raise NotImplementedError()

    def restart(self, services, yes):
        """Starts services

        Args:
          * services: a list with the services name to start
        """
        raise NotImplementedError()

    def execute(self, container_name, *args):
        """Execute a command on a container

        Args:
          * container_name: the name of the container
          * *args: the command, like 'ps', 'aux'
        """
        raise NotImplementedError()

    def cmd_up(self, services=None, verbose=False, build_log=None):
        """Start the stack"""
        raise NotImplementedError()

    def cmd_ps(self):
        """Hook for drivers to provide a listing process feature"""
        raise NotImplementedError()

    def cmd_status(self):
        """Hook for drivers to provide a check process feature"""
        raise NotImplementedError()

    # pylint: disable=unused-argument
    def clean(self, kill=False):
        """Hook for drivers to provide a clean/stop process feature"""
        raise NotImplementedError()

    def encrypt(self, password=None):
        """Hook for drivers to provide a kard encrypt feature"""
        raise NotImplementedError()

    def decrypt(self, password=None):
        """Hook for drivers to provide a kard decrypt feature"""
        raise NotImplementedError()


class BaseDriver(AbstractDriver, ABC):
    metas = []

    def get_templates(self):
        templates = []
        templates_path = self.kard.env.pkr_path / self.kard.env.template_dir

        # Process templates
        for container in self.kard.env.get_container():
            for template in self.kard.env.get_container(container)["templates"]:
                templates.append(
                    {
                        "source": templates_path / template,
                        "origin": templates_path,
                        "destination": "",
                        "subfolder": "templated",
                    }
                )

        # Process requirements
        for src in self.kard.env.get_requires():
            templates.append(
                {
                    "source": src["origin"],
                    "origin": src["origin"],
                    "destination": src["dst"],
                    "subfolder": "templated",
                    "excluded_paths": src.get("exclude", []),
                    "gen_template": False,
                }
            )

        return templates



# Copyright© 1986-2024 Altair Engineering Inc.

"""Git extension to allow using sources from volumes or build them into the
image"""
import os

import jinja2

from pkr.kard import Kard
from . import ExtMixin


class AutoVolume(ExtMixin):
    """Macro for dockerfiles templating which allows either mounting or
    adding files according to the meta `use_volume`.
    """

    @staticmethod
    def get_context_template_data(password=None):
        kard = Kard.load_current(None, password)
        return {"add_file": add_file, "use_volume": kard.env.get("use_volume", False)}


@jinja2.pass_context
def add_file(context, paths):
    """This function is used inside the dockerfiles templates to
    render them by using either the ADD or VOLUME instruction."""
    if context["use_volume"]:
        paths = paths.get("common", {})
        paths = [f'"{path}"' for path in list(paths.keys())]
        if len(paths):
            return f"VOLUME [{', '.join(paths)}]"
        return ""

    lines = []
    paths = dict(list(paths.get("common", {}).items()) + list(paths.get("copied", {}).items()))
    for remote, local in sorted(paths.items()):
        lines.append(f'ADD "{local}" "{remote}"')
    return os.linesep.join(lines)



# Copyright© 1986-2024 Altair Engineering Inc.

# pylint: disable=C0111,E1101,R0912,R0913

"""Utils functions for pkr"""

import hashlib
from builtins import input
from builtins import range
from builtins import str
from enum import Enum
from fnmatch import fnmatch
from glob import glob
import json
import os
from pathlib import Path
import re
import secrets
import shutil
import time
import platform

from Crypto.Cipher import AES
from Crypto.Hash import SHA256
from Crypto import Random
import docker
import jinja2
from passlib.apache import HtpasswdFile

ENV_FOLDER = "env"
KARD_FOLDER = "kard"
PATH_ENV_VAR = "PKR_PATH"


class Cmd(Enum):
    OTHER = 0
    ENCRYPT = 1
    DECRYPT = 2


class PkrException(Exception):
    """pkr Exception"""


class KardInitializationException(PkrException):
    """pkr Exception"""


class PasswordException(PkrException):
    def __init__(
        self, message="Encryption password is not specified (use global -p <password> option)"
    ):
        super().__init__(message)


# pylint: disable=C0415
def is_pkr_path(path):
    """Check environments files to deduce if path is a usable pkr path"""
    return path.is_dir() and len(list(path.glob(f"{ENV_FOLDER}/*/env.yml"))) > 0


def get_pkr_path(raise_if_not_found=True):
    """Return the path of the pkr folder

    If the env. var 'PKR_PATH' is specified, it is returned, otherwise a
    KeyError exception is raised.
    """

    full_path = Path(os.environ.get(PATH_ENV_VAR, os.getcwd())).absolute()
    pkr_path = full_path
    while pkr_path.parent != pkr_path:
        if is_pkr_path(pkr_path):
            return pkr_path
        pkr_path = pkr_path.parent

    if raise_if_not_found and not is_pkr_path(pkr_path):
        raise KardInitializationException(
            f"{'Given' if PATH_ENV_VAR in os.environ else 'Current'} path {full_path} is not a "
            f"valid pkr path, no usable env found"
        )

    return pkr_path


def get_kard_root_path():
    """Return the root path of Kards"""
    return get_pkr_path() / KARD_FOLDER


def get_timestamp():
    """Return a string timestamp"""
    return time.strftime("%Y%m%d-%H%M%S")


class HashableDict(dict):
    """Extends dict with a __hash__ method to make it unique in a set"""

    def __key(self):
        return json.dumps(self)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return self.__key() == other.__key()  # pylint: disable=W0212


def merge(source, destination, overwrite=True):
    """Deep merge 2 dicts

    Warning: the source dict is merged INTO the destination one. Make a copy
    before using it if you do not want to destroy the destination dict.
    """
    if not source:
        return destination
    if destination is None:
        destination = {}
    for key, value in list(source.items()):
        if isinstance(value, dict):
            # Handle type mismatch
            if overwrite and not isinstance(destination.get(key), dict):
                destination[key] = {}
            # get node or create one
            node = destination.setdefault(key, {})
            merge(value, node, overwrite)
        elif isinstance(value, list):
            if key in destination:
                # Handle type mismatch
                if overwrite and not isinstance(destination[key], list):
                    destination[key] = []
                try:
                    destination[key] = list(dict.fromkeys(destination[key] + value))
                # Prevent errors when having unhashable dict types
                except TypeError:
                    destination[key].extend(value)
            else:
                destination[key] = value
        elif overwrite or key not in destination:
            destination[key] = value

    return destination


def diff(previous, current):
    """Deep diff 2 dicts

    Return a dict of new elements in the `current` dict.
    Does not handle removed elements.
    """
    result = {}

    for key in current.keys():
        value = current[key]
        p_value = previous.get(key, None)
        if p_value is None:
            result[key] = value
            continue

        if isinstance(value, dict):
            difference = diff(previous.get(key, {}), value)
            if difference:
                result[key] = difference

        elif isinstance(value, list):
            value = [x for x in value if x not in p_value]
            if value != []:
                result[key] = value

        elif value != p_value:
            result[key] = value

    return result


def generate_password(pw_len=15):
    """Generate a password"""
    alphabet = "abcdefghijklmnopqrstuvwxyz"
    upperalphabet = alphabet.upper()
    pwlist = []

    for _ in range(pw_len // 3):
        pwlist.append(secrets.choice(alphabet))
        pwlist.append(secrets.choice(upperalphabet))
        pwlist.append(str(secrets.randbelow(10)))
    for _ in range(pw_len - len(pwlist)):
        pwlist.append(secrets.choice(alphabet))

    secrets.SystemRandom().shuffle(pwlist)
    return "".join(pwlist)


def ask_input(name):
    return input(f"Missing meta({name}):")


class TemplateEngine:
    def __init__(self, tpl_context):
        """Init templating context (filters and functions)"""
        self.tpl_context = tpl_context.copy()

        self.pkr_path = get_pkr_path()
        self.tpl_env = jinja2.Environment(
            extensions=["jinja2_ansible_filters.AnsibleCoreFiltersExtension"],
            loader=jinja2.FileSystemLoader(str(self.pkr_path)),
        )

        def sha256(string):
            return hashlib.sha256(string.encode("utf-8")).hexdigest()

        self.tpl_env.filters["sha256"] = sha256

        def format_htpasswd(username, password):
            ht = HtpasswdFile()
            ht.set_password(username, password)
            return ht.to_string().rstrip().decode("utf-8")

        self.tpl_context["format_htpasswd"] = format_htpasswd

    def process_template(self, template_file):
        """Process a template and render it in the context

        Args:
          - template_file: the template to load

        Return the result of the processed template.
        """

        rel_template_file = str(template_file.relative_to(self.pkr_path))

        template = self.tpl_env.get_template(rel_template_file)
        out = template.render(self.tpl_context)
        return out

    def process_string(self, string):
        """Process a string and render it in the context

        Args:
          - template_file: the string to template

        Return the result of the processed string.
        """
        template = self.tpl_env.from_string(string)
        out = template.render(self.tpl_context)
        return out

    def copy(self, path, origin, local_dst, excluded_paths, gen_template=True):
        """Copy a tree recursively, while excluding specified files

        Args:
          - path: the file or folder to copy
          - origin: the base folder of all paths
          - local_dst: the destination folder / file
          - excluded_paths: the list of unwanted excluded files
        """
        path_str = str(path)
        if "*" in path_str:
            file_list = [Path(p) for p in glob(path_str)]
            if "*" in origin.name:
                origin = origin.parent

            for path_it in file_list:
                rel_local_dst = path_it.relative_to(origin)
                full_local_dst = local_dst / rel_local_dst

                self.copy(path_it, path_it, full_local_dst, excluded_paths, gen_template)
        elif path.is_file():
            # Direct match for excluded paths
            if path in excluded_paths:
                return
            if path != origin:
                # path = /pkr/src/backend/api/__init__.py
                abs_path = path.relative_to(origin)
                # path = api/__init__.py
                dst_path = local_dst / abs_path
                # path = docker-context/backend/api/__init__.py
            else:
                # Here we avoid having a '.' as our abs_path
                dst_path = local_dst
            # We ensure that the containing folder exists

            dst_path.parent.mkdir(parents=True, exist_ok=True)
            if gen_template and path.name.endswith(".template"):
                # If the dst_local contains the filename with template
                if not dst_path.is_dir():
                    if dst_path.name.endswith(".template"):
                        dst_path = self.remove_ext(dst_path)
                else:  # We create the destination path
                    dst_path = dst_path / self.remove_ext(path.name)
                out = self.process_template(path)
                dst_path.write_text(out)
                shutil.copystat(path_str, str(dst_path))
                # os.chmod(str(dst_path), 0o600) # make invisible to the world
            else:
                shutil.copy2(path_str, str(dst_path))
        elif path.is_dir():
            for path_it in path.iterdir():
                path_it = path / path_it
                if not any((fnmatch(str(path_it), str(exc_path)) for exc_path in excluded_paths)):
                    self.copy(path_it, origin, local_dst, excluded_paths, gen_template)

    @staticmethod
    def remove_ext(path):
        """Remove the portion of a string after the last dot."""
        return path.parent / path.stem


FLAGS = re.VERBOSE | re.MULTILINE | re.DOTALL
WHITESPACE = re.compile(r"[ \t\n\r]*", FLAGS)


class ConcatJSONDecoder(json.JSONDecoder):
    def decode(self, s, _w=WHITESPACE.match):
        s_len = len(s)

        objs = []
        end = 0
        while end != s_len:
            obj, end = self.raw_decode(s, idx=_w(s, end).end())
            end = _w(s, end).end()
            objs.append(obj)
        return objs


def get_current_container():
    """Return container inspect if we run in docker, None otherwise"""
    if platform.system() != "Linux":
        return None
    with Path("/proc/self/cgroup").open(encoding="utf-8") as cgroup_file:
        for line in cgroup_file:
            if "docker" in line:
                container_id = line[line.rindex("/") + 1 :].strip()
                md = re.match(r"docker-(.+)\.scope$", container_id)
                if md:
                    container_id = md.group(1)
                cli = docker.DockerClient(version="auto")  # Default to /var/run/docker.sock
                return cli.containers.get(container_id)
    return None


def ensure_key_present(key, default, data, path=None):
    """Ensure that a key is present, set the default is present, or ask
    the user to input it."""

    if key in data:
        return data.get(key)
    if key in default:
        return default.get(key)

    return ask_input((path or "") + key)


def ensure_definition_matches(definition, defaults, data, path=None):
    """Recursive function that ensures data is provided.

    Ask for it if not.
    """
    path = path or ""
    if isinstance(definition, dict):
        values = {
            k: ensure_definition_matches(
                definition=v,
                defaults=defaults.get(k, []),
                data=data.get(k, {}),
                path=path + k + "/",
            )
            for k, v in definition.items()
            if v is not None
        }
        values.update(
            {
                k: ensure_key_present(k, defaults, data, path)
                for k, v in definition.items()
                if v is None
            }
        )
        return values

    if isinstance(definition, list):
        values = {}
        for element in definition:
            values.update(
                ensure_definition_matches(
                    definition=element, defaults=defaults, data=data, path=path
                )
            )
        return values

    value = ensure_key_present(definition, defaults, data, path)
    return {definition: value}


def create_pkr_folder(pkr_path=None):
    """Creates a folder structure for pkr.

    This looks like:
    PKR_PATH/
    ├── env/
    │   └── dev/
    │       └── env.yaml
    └── kard/
    """
    pkr_path = pkr_path or get_pkr_path(False)

    (pkr_path / "env" / "dev").mkdir(parents=True)
    (pkr_path / "env" / "dev" / "env.yml").touch()
    (pkr_path / "kard").mkdir(parents=True)


def dedup_list(src):
    """Dedup src list (in-place) and yield duplicates"""
    for item in set(src):
        if src.count(item) != 1:
            yield item
            src.remove(item)


def merge_lists(src, dest, insert=True):
    """Merge lists avoiding duplicates"""
    if insert:
        for x in reversed(src):
            if x in dest:
                continue
            dest.insert(0, x)
    else:
        dest.extend([x for x in src if x not in dest])
    return dest


def encrypt_swap(file, file_enc, password):
    enc = encrypt_file(file, password)
    with file_enc.open("wb") as fe:
        os.chmod(file_enc, 0o600)
        fe.write(enc)
        file.unlink()


def encrypt_file(file, password=None):
    if password is None:
        raise PasswordException()
    with file.open("rb") as fl:
        data = fl.read()
        data_enc = encrypt_with_key(password.encode("utf-8"), data)
        return data_enc


def decrypt_swap(file, file_enc, password):
    data = decrypt_file(file_enc, password)
    file.write_text(data.decode("utf-8"))
    os.chmod(file, 0o600)
    file_enc.unlink()


def decrypt_file(file, password=None):
    if password is None:
        raise PasswordException()
    with file.open("rb") as fl:
        data_enc = fl.read()
        data = decrypt_with_key(password.encode("utf-8"), data_enc)
        return data


def encrypt_with_key(key: bytes, source: bytes) -> bytes:
    key = SHA256.new(key).digest()  # use SHA-256 over our key to get a proper-sized AES key
    i_v = Random.new().read(AES.block_size)  # generate i_v
    encryptor = AES.new(key, AES.MODE_CBC, i_v)
    padding = AES.block_size - len(source) % AES.block_size  # calculate needed padding
    src = b"".join([source, bytes([padding]) * padding])
    data = i_v + encryptor.encrypt(src)  # store the i_v at the beginning and encrypt
    return data


def decrypt_with_key(key: bytes, source: bytes) -> bytes:
    key = SHA256.new(key).digest()  # use SHA-256 over our key to get a proper-sized AES key
    i_v = source[: AES.block_size]  # extract the i_v from the beginning
    decryptor = AES.new(key, AES.MODE_CBC, i_v)
    data = decryptor.decrypt(source[AES.block_size :])  # decrypt
    padding = data[-1]  # pick the padding value from the end
    if data[-padding:] != bytes([padding]) * padding:
        raise Exception("Incorrect decryption password")
    return data[:-padding]  # remove the padding



# Copyright© 1986-2024 Altair Engineering Inc.

"""Extension to generate additional files as templates."""

from pkr.ext import ExtMixin
from pkr.kard import Kard
from pkr.utils import get_pkr_path


class BasicTemplate(ExtMixin):
    """pkr extension to configure generate extra files using templates.

    It is invoked during the make call.
    """

    @staticmethod
    def populate_kard():
        """Generate the template listed in the env."""

        kard = Kard.load_current()

        extra_templates = kard.env.get("templates", [])

        tpl_engine = kard.get_template_engine()

        for tpl in extra_templates:
            tpl_path = get_pkr_path() / tpl["template"]
            # Render the template
            tpl_render = tpl_engine.process_template(tpl_path)

            # Write it out
            dst = kard.path / tpl["dst"]
            dst.write_text(tpl_render)



# Copyright© 1986-2024 Altair Engineering Inc.

from pkr.kard import Kard
from .utils import pkrTestCase


class TestExt(pkrTestCase):
    pkr_folder = "ext"

    def test_kard_features_order_is_deterministic(self):
        res = []
        for i in range(3):
            with open(str(self.env_test.path / "meta1.yml"), "r") as f:
                kard = Kard.create(
                    name="test", env="test", driver="compose", features="a,b", extra={}, meta=f
                )
            res.append((kard.meta["features"], kard.extensions.extensions))
            self.assertEqual(repr(res[-1]), repr(res[0]))
        self.assertEqual(res[0][0], ["auto-volume", "a", "ext_mock", "b"])

    def test_ext_loaded_from_pkr_path(self):
        with open(str(self.env_test.path / "meta1.yml"), "r") as f:
            kard = Kard.create(
                name="test", env="test", driver="compose", features="a,b", extra={}, meta=f
            )
        self.assertIn("ext_mock", kard.extensions.extensions)
        self.assertIn({"test": "Ok"}, kard.extensions.get_context_template_data())

    def test_ext_loaded_from_entrypoints_group_pkr_extensions(self):
        with open(str(self.env_test.path / "meta1.yml"), "r") as f:
            kard = Kard.create(
                name="test", env="test", driver="compose", features="a,b", extra={}, meta=f
            )
        self.assertIn("auto-volume", kard.extensions.extensions)
        self.assertIn("use_volume", kard.extensions.get_context_template_data()[0])
